"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypeAlias, TypedDict

from typing_extensions import NotRequired, Required, TypeVar

from .http import ResponseDict
from .id import ID32

if TYPE_CHECKING:
    from ..abc import PartialUser
    from ..enums import Type
    from ..id import ID
    from ..user import ClientUser, User as User_
    from .id import ID64


class User(TypedDict):
    steamid: str
    personaname: str
    primaryclanid: str
    profileurl: str
    realname: str
    # enums
    communityvisibilitystate: int
    profilestate: int
    commentpermission: int
    personastate: int
    personastateflags: int
    # avatar
    avatar: str
    avatarmedium: str
    avatarfull: str
    avatarhash: str
    # country stuff
    loccountrycode: str
    locstatecode: int
    loccityid: int
    # game stuff
    gameid: str  # game app id
    gameextrainfo: str  # game name
    # unix timestamps
    timecreated: int
    lastlogoff: int


class InventoryInfo(TypedDict):
    appid: int
    name: str
    inventory_logo: str
    icon: str
    asset_count: int
    trade_permissions: str
    load_failed: int
    store_vetted: int
    owner_only: bool
    rgContexts: dict[Any, InventoryInfoContexts]


class InventoryInfoContexts(TypedDict):
    id: int
    name: str
    asset_count: int


class RawUserBan(TypedDict):
    SteamId: str
    CommunityBanned: bool
    VACBanned: bool
    NumberOfVACBans: int
    DaysSinceLastBan: int
    NumberOfGameBans: int
    EconomyBan: str


class UserBan(TypedDict):  # steam does such a horrific job exposing this I recreate the dict with lowered keys
    steamid: ID64
    community_banned: bool
    vac_banned: bool
    number_of_vac_bans: int
    days_since_last_ban: int
    number_of_game_bans: int
    economy_ban: str


class UserBadges(TypedDict):
    badges: list[UserBadgeBadge]
    player_xp: int
    player_level: int
    player_xp_needed_to_level_up: int
    player_xp_needed_current_level: int


class UserBadgeBadge(TypedDict):
    badgeid: int
    appid: NotRequired[int]
    level: int
    completion_time: int
    xp: int
    communityitemid: NotRequired[int]
    border_color: NotRequired[int]
    scarcity: int


class CommunityBadgeProgressQuest(TypedDict):
    questid: int
    completed: bool


class TradeHoldDurations(TypedDict):
    my_escrow: NotRequired[Escrow]
    their_escrow: NotRequired[Escrow]
    both_escrow: NotRequired[Escrow]


class Escrow(TypedDict, total=False):
    escrow_end_duration_seconds: Required[int]
    escrow_end_date: int
    escrow_end_date_rfc3339: str


class AuthenticateUserTicketParams(TypedDict):
    result: str
    steamid: str
    ownersteamid: str
    vacbanned: bool
    publisherbanned: bool


GetUserGroupList: TypeAlias = ResponseDict[dict[Literal["groups"], list[dict[Literal["gid"], ID32]]]]
GetPlayerBans: TypeAlias = dict[Literal["players"], list[RawUserBan]]
FriendsList: TypeAlias = dict[Literal["friendslist"], dict[Literal["friends"], list[dict[Literal["steamid"], str]]]]
AuthenticateUserTicket: TypeAlias = dict[Literal["params"], AuthenticateUserTicketParams]
GetSteamLevel: TypeAlias = ResponseDict[dict[Literal["player_level"], int]]

UserT = TypeVar("UserT", bound="PartialUser", default="PartialUser", covariant=True)  # for use when just from cache
Author: TypeAlias = "User_ | ClientUser | PartialUser"
AuthorT = TypeVar(
    "AuthorT", bound="PartialUser", default=Author, covariant=True
)  # for use when something comes from _maybe_user
IndividualID: TypeAlias = "ID[Literal[Type.Individual]]"



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_twofactor.steamclient.proto
# plugin: python-betterproto
# Last updated 09/09/2021

import betterproto

from .msg import UnifiedMessage


class StatusRequest(UnifiedMessage, um_name="TwoFactor.QueryStatus"):
    steamid: int = betterproto.fixed64_field(1)


class StatusResponse(UnifiedMessage, um_name="TwoFactor.QueryStatus"):
    state: int = betterproto.uint32_field(1)
    inactivation_reason: int = betterproto.uint32_field(2)
    authenticator_type: int = betterproto.uint32_field(3)
    authenticator_allowed: bool = betterproto.bool_field(4)
    steamguard_scheme: int = betterproto.uint32_field(5)
    token_gid: str = betterproto.string_field(6)
    email_validated: bool = betterproto.bool_field(7)
    device_identifier: str = betterproto.string_field(8)
    time_created: int = betterproto.uint32_field(9)
    revocation_attempts_remaining: int = betterproto.uint32_field(10)
    classified_agent: str = betterproto.string_field(11)
    allow_external_authenticator: bool = betterproto.bool_field(12)
    time_transferred: int = betterproto.uint32_field(13)


class AddAuthenticatorRequest(UnifiedMessage, um_name="TwoFactor.AddAuthenticator"):
    steamid: int = betterproto.fixed64_field(1)
    authenticator_time: int = betterproto.uint64_field(2)
    serial_number: int = betterproto.fixed64_field(3)
    authenticator_type: int = betterproto.uint32_field(4)
    device_identifier: str = betterproto.string_field(5)
    sms_phone_id: str = betterproto.string_field(6)
    http_headers: list[str] = betterproto.string_field(7)


class AddAuthenticatorResponse(UnifiedMessage, um_name="TwoFactor.AddAuthenticator"):
    shared_secret: bytes = betterproto.bytes_field(1)
    serial_number: int = betterproto.fixed64_field(2)
    revocation_code: str = betterproto.string_field(3)
    uri: str = betterproto.string_field(4)
    server_time: int = betterproto.uint64_field(5)
    account_name: str = betterproto.string_field(6)
    token_gid: str = betterproto.string_field(7)
    identity_secret: bytes = betterproto.bytes_field(8)
    secret_1: bytes = betterproto.bytes_field(9)
    status: int = betterproto.int32_field(10)


class SendEmailRequest(UnifiedMessage, um_name="TwoFactor.SendEmail"):
    steamid: int = betterproto.fixed64_field(1)
    email_type: int = betterproto.uint32_field(2)
    include_activation_code: bool = betterproto.bool_field(3)


class SendEmailResponse(UnifiedMessage, um_name="TwoFactor.SendEmail"):
    pass


class FinalizeAddAuthenticatorRequest(UnifiedMessage, um_name="TwoFactor.FinalizeAddAuthenticator"):
    steamid: int = betterproto.fixed64_field(1)
    authenticator_code: str = betterproto.string_field(2)
    authenticator_time: int = betterproto.uint64_field(3)
    activation_code: str = betterproto.string_field(4)
    http_headers: list[str] = betterproto.string_field(5)


class FinalizeAddAuthenticatorResponse(UnifiedMessage, um_name="TwoFactor.FinalizeAddAuthenticator"):
    success: bool = betterproto.bool_field(1)
    want_more: bool = betterproto.bool_field(2)
    server_time: int = betterproto.uint64_field(3)
    status: int = betterproto.int32_field(4)


class RemoveAuthenticatorRequest(UnifiedMessage, um_name="TwoFactor.RemoveAuthenticator"):
    revocation_code: str = betterproto.string_field(2)
    revocation_reason: int = betterproto.uint32_field(5)
    steamguard_scheme: int = betterproto.uint32_field(6)
    remove_all_steamguard_cookies: bool = betterproto.bool_field(7)


class RemoveAuthenticatorResponse(UnifiedMessage, um_name="TwoFactor.RemoveAuthenticator"):
    success: bool = betterproto.bool_field(1)
    server_time: int = betterproto.uint64_field(3)
    revocation_attempts_remaining: int = betterproto.uint32_field(5)


class CreateEmergencyCodesRequest(UnifiedMessage, um_name="TwoFactor.CreateEmergencyCodes"):
    code: str = betterproto.string_field(1)


class CreateEmergencyCodesResponse(UnifiedMessage, um_name="TwoFactor.CreateEmergencyCodes"):
    codes: list[str] = betterproto.string_field(1)


class DestroyEmergencyCodesRequest(UnifiedMessage, um_name="TwoFactor.DestroyEmergencyCodes"):
    steamid: int = betterproto.fixed64_field(1)


class DestroyEmergencyCodesResponse(UnifiedMessage, um_name="TwoFactor.DestroyEmergencyCodes"):
    pass


class ValidateTokenRequest(UnifiedMessage, um_name="TwoFactor.ValidateToken"):
    code: str = betterproto.string_field(1)


class ValidateTokenResponse(UnifiedMessage, um_name="TwoFactor.ValidateToken"):
    valid: bool = betterproto.bool_field(1)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import abc
import re
from collections.abc import Callable
from contextlib import nullcontext
from types import GenericAlias
from typing import TYPE_CHECKING, Final, Generic, Literal, cast

import aiohttp
from typing_extensions import TypeVar

from ._const import JSON_LOADS, URL
from .enums import Instance, Type, TypeChar, Universe
from .errors import InvalidID
from .types.id import ID32, ID64, Intable

if TYPE_CHECKING:
    from aiohttp import ClientSession

    from .types.http import StrOrURL
    from .types.user import IndividualID


__all__ = ("ID",)


def parse_id64(
    id: Intable,
    /,
    *,
    type: Type | None = None,
    universe: Universe | None = None,
    instance: Instance | None = None,
) -> ID64:
    """Convert various representations of Steam IDs to its Steam 64-bit ID.

    Parameters
    ----------
    id
        The ID to convert.
    type
        The type of the ID.
    universe
        The universe of the ID.
    instance
        The instance of the ID.

    Examples
    --------
    .. code:: python

        parse_id64(12345)
        parse_id64("12345")  # account ids
        parse_id64(12345, type=steam.Type.Clan)  # makes what would be interpreted as a user id into a clan id64
        parse_id64(103582791429521412)
        parse_id64("103582791429521412")  # id64s
        parse_id64("STEAM_1:0:2")  # id2
        parse_id64("[g:1:4]")  # id3

    Raises
    ------
    :exc:`.InvalidID`
        The created 64-bit Steam ID would be invalid.

    Returns
    -------
    The 64-bit Steam ID.
    """

    if not id and type is None and universe is None and instance is None:
        return ID64(0)

    try:
        id = int(id)
    except ValueError:
        # textual input e.g. [g:1:4]
        if not isinstance(id, str):
            raise InvalidID(id, type, universe, instance, "it cannot be parsed as an int or str") from None
        result = ID.from_id2(id) or ID.from_id3(id)
        if result is None:
            raise InvalidID(id, type, universe, instance, "it cannot be parsed") from None
        return result.id64
    else:
        # numeric input
        if id < 0:
            raise InvalidID(id, type, universe, instance, "it is too small")
        elif 0 <= id < 2**32:  # 32 bit
            type = type or Type.Individual
            universe = universe or Universe.Public

            if instance is None:
                instance = Instance.Desktop if type in (Type.Individual, Type.GameServer) else Instance.All

            if not (0 <= universe < 1 << 8):
                raise InvalidID(id, type, universe, instance, "universe is bigger than 8 bits")
            if not (0 <= type < 1 << 4):
                raise InvalidID(id, type, universe, instance, "type is bigger than 4 bits")
            if not (0 <= instance < 1 << 20):
                raise InvalidID(id, type, universe, instance, "instance is bigger than 20 bits")
        elif id < 2**64:  # 64 bit
            universe = Universe.try_value((id >> 56) & 0xFF)
            type = Type.try_value((id >> 52) & 0xF)
            instance = Instance.try_value((id >> 32) & 0xFFFFF)
            id &= 0xFFFFFFFF
        else:
            raise InvalidID(id, type, universe, instance, "it is too large")

    return ID64(universe << 56 | type << 52 | instance << 32 | id)


ID2_REGEX = re.compile(r"STEAM_(?P<universe>\d+):(?P<remainder>[0-1]):(?P<id>\d{1,10})")
ID3_REGEX = re.compile(
    (
        rf"\[(?P<type>[i{''.join(TypeChar._member_map_)}]):"
        rf"(?P<universe>[{min(Universe).value}-{max(Universe).value}]):"
        r"(?P<id>[0-9]{1,10})"
        r"(:(?P<instance>\d+))?]"
    )
)


_INVITE_HEX = "0123456789abcdef"
_INVITE_CUSTOM = "bcdfghjkmnpqrtvw"
_INVITE_VALID = f"{_INVITE_HEX}{_INVITE_CUSTOM}"
_URL_START = r"(?:https?://)?(?:www\.)?"
INVITE_REGEX = re.compile(rf"(?:{_URL_START}(?:s\.team/p/))?(?P<code>[\-{_INVITE_VALID}]{{1,8}})")


def _invite_custom_sub(
    s: str,
    repl: Callable[[re.Match[str]], str] = lambda m, map=dict(zip(_INVITE_CUSTOM, _INVITE_HEX)): map[m.group()],
    pattern: re.Pattern[str] = re.compile(f"[{_INVITE_CUSTOM}]"),
    /,
) -> str:
    return pattern.sub(repl, s)


def _invite_hex_sub(
    s: str,
    repl: Callable[[re.Match[str]], str] = lambda m, map=dict(zip(_INVITE_HEX, _INVITE_CUSTOM)): map[m.group()],
    pattern: re.Pattern[str] = re.compile(f"[{_INVITE_HEX}]"),
    /,
) -> str:
    return pattern.sub(repl, s)


USER_URL_PATHS = frozenset({"id", "profiles", "user"})
CLAN_URL_PATHS = frozenset({"gid", "groups", "app", "games"})
URL_REGEX = re.compile(
    rf"{_URL_START}(?P<clean_url>steamcommunity\.com/(?P<type>{'|'.join(USER_URL_PATHS | CLAN_URL_PATHS)})/[^/]*)"
)
USER_ID64_FROM_URL_REGEX = re.compile(r"g_rgProfileData\s*=\s*(?P<json>{.*?});\s*")
CLAN_ID64_FROM_URL_REGEX = re.compile(r"OpenGroupChat\(\s*'(?P<steamid>\d+)'\s*\)")


async def id64_from_url(url: StrOrURL, /, session: aiohttp.ClientSession | None = None) -> ID64 | None:
    """Fetches the 64-bit Steam ID from a Steam Community URL or ``None``.

    See Also
    --------
    :meth:`ID.from_url` for more information.
    """
    id = await ID.from_url(url, session=session)
    return id.id64 if id is not None else None


_ID64_TO_ID32: Final = cast(Callable[[int], ID32], 0xFFFFFFFF.__and__)


TypeT = TypeVar("TypeT", bound=Type, default=Type, covariant=True)


class ID(Generic[TypeT], metaclass=abc.ABCMeta):
    """Convert a Steam ID between its various representations.

    .. container:: operations

        .. describe:: x == y

            Checks if two IDs are equal.

        .. describe:: hash(x)

            Returns the hash of the ID.

        .. describe:: str(x)

            Returns the string representation of :attr:`id64`.

        .. describe:: int(x)

            Returns the :attr:`id64` of the ID.

        .. describe:: format(x, format_spec)

            Formats the ID using the given format spec.

            Prefixes of ``32``, ``64`` can be used to specify which of :attr:`id` or :attr:`id64` to use.
            Anything after the prefix is passed to :func:`format`.

            E.g.

            .. code-block:: pycon

                >>> format(steam_id, "64x")  # formats the `id64` as a hex string
                "11000010264339c"
                >>> format(steam_id, "32b")  # formats the `id` as binary
                "10011001000011001110011100"
                >>> f"{steam_id:32b}"  # same as above


    Parameters
    ----------
    id
        The ID to convert.
    type
        The type of the ID.
    universe
        The universe of the ID.
    instance
        The instance of the ID.
    """

    # format of a 64-bit steam ID:
    # 0b0000000100010000000000000000000100010001001001110100110011000010
    #   └───┰──┘└─┰┘└─────────┰────────┘└──────────────┰───────────────┘
    #       │     │           │                        │
    #   universe  └ type      └ instance               └ account id
    #   (8 bits)    (4 bits)    (20 bits)                (32 bits)
    #   Universe    Type        InstanceFlag             ID32
    #   Public      Individual  All                      287788226

    __slots__ = ("id64", "__weakref__")
    __class_getitem__ = classmethod(
        GenericAlias
    )  # want the different behaviour between typing._GenericAlias to do with attribute forwarding

    def __init__(
        self,
        id: Intable,
        *,
        type: TypeT | None = None,
        universe: Universe | None = None,
        instance: Instance | None = None,
    ):
        self.id64: Final = parse_id64(id, type=type, universe=universe, instance=instance)
        """The Steam ID's 64-bit ID."""

    def __int__(self) -> ID64:
        return self.id64

    def __eq__(self, other: object) -> bool:
        return isinstance(other, ID) and self.id64 == other.id64

    def __str__(self) -> str:
        return str(self.id64)

    def __hash__(self) -> int:
        return hash(self.id64)

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}(id={self.id}, type={self.type!r}, universe={self.universe!r}, instance={self.instance!r})"

    def __format__(self, format_spec: str, /) -> str:
        match format_spec[:2]:
            case "64" | "":
                return format(self.id64, format_spec[2:])
            case "32":
                return format(self.id, format_spec[2:])
            case _:
                raise ValueError(f"Unknown format specifier {format_spec!r}")

    @property
    def universe(self) -> Universe:
        """The Steam universe of the ID."""
        return Universe.try_value((self.id64 >> 56) & 0xFF)

    @property
    def type(self) -> TypeT:
        """The Steam type of the ID."""
        return cast(TypeT, Type.try_value((self.id64 >> 52) & 0xF))

    @property
    def instance(self) -> Instance:
        """The instance of the ID."""
        return Instance.try_value((self.id64 >> 32) & 0xFFFFF)

    @property
    def id(self, _ID64_TO_ID32: Callable[[ID64], ID32] = _ID64_TO_ID32, /) -> ID32:
        """The Steam ID's 32-bit ID."""
        return _ID64_TO_ID32(self.id64)

    @property
    def id2(self) -> str:
        """The Steam ID's ID 2.

        e.g ``STEAM_1:0:1234``.
        """
        return f"STEAM_{self.universe.value}:{self.id % 2}:{self.id >> 1}"

    @property
    def id2_zero(self) -> str:
        """The Steam ID's ID 2 accounted for bugged GoldSrc and Orange Box games.

        Note
        ----
        In these games the accounts :attr:`universe`, ``1`` for :class:`.Type.Public`, should be the ``X`` component of
        ``STEAM_X:0:1234`` however, this was bugged and the value of ``X`` was ``0``.

        e.g ``STEAM_0:0:1234``.
        """
        return self.id2.replace("_1", "_0")

    @property
    def id3(self) -> str:
        """The Steam ID's ID 3.

        e.g ``[U:1:1234]``.
        """
        type_char = TypeChar(self.type).name
        instance = None

        match self.type:
            case Type.AnonGameServer | Type.Multiseat:
                instance = self.instance
            case Type.Individual:
                if self.instance != Instance.Desktop:
                    instance = self.instance
            case Type.Chat:
                if self.instance & Instance.ChatClan > 0:
                    type_char = "c"
                elif self.instance & Instance.ChatLobby > 0:
                    type_char = "L"
                else:
                    type_char = "T"

        return f"[{type_char}:{self.universe.value}:{self.id}{f':{instance.value}' if instance is not None else ''}]"

    # @property
    # @overload
    # def invite_code(self: ID[Type.Individual]) -> str:
    #     ...
    #
    # @property
    # @overload
    # def invite_code(self: ID[~Type.Individual]) -> None:
    #     ...

    @property
    def invite_code(self) -> str | None:
        """The Steam ID's invite code in the s.team invite code format.

        e.g. ``cv-dgb``.
        """
        if self.type == Type.Individual and self.is_valid():
            invite_code = _invite_hex_sub(f"{self:32x}")
            split_idx = len(invite_code) // 2
            return invite_code if split_idx == 0 else f"{invite_code[:split_idx]}-{invite_code[split_idx:]}"

    # @property
    # @overload
    # def invite_url(self: ID[Type.Individual]) -> str:
    #     ...
    #
    # @property
    # @overload
    # def invite_url(self: ID[~Type.Individual]) -> None:
    #     ...
    #

    @property
    def invite_url(self) -> str | None:
        """The Steam ID's full invite code URL.

        e.g ``https://s.team/p/cv-dgb``.
        """
        code = self.invite_code
        return f"https://s.team/p/{code}" if code else None

    # @property
    # @overload
    # def community_url(self: ID[Type.Individual | Type.Clan]) -> str:
    #     ...
    #
    # @property
    # @overload
    # def community_url(self: ID[~(Type.Individual | Type.Clan)]) -> None:
    #     ...

    @property
    def community_url(self) -> str | None:
        """The Steam ID's community url if it is a :attr:`.Type.Individual` or :attr:`.Type.Clan`.

        e.g ``https://steamcommunity.com/profiles/123456789`` or ``https://steamcommunity.com/gid/123456789``.
        """
        match self.type:
            case Type.Individual:
                return str(URL.COMMUNITY / f"profiles/{self.id64}")
            case Type.Clan:
                return str(URL.COMMUNITY / f"gid/{self.id64}")
            case _:
                return None

    def is_valid(self) -> bool:
        """Whether this Steam ID is valid.

        A Steam ID is currently considered valid if:

            - It is in ``(0, 2**64)``
            - :attr:`universe` is in ``(Invalid, Dev]``
            - :attr:`type` is in ``(Invalid, AnonUser]``
            - If :attr:`type` is :attr:`.Type.Individual`:
                - :attr:`id` is non-zero
                - :attr:`instance` is in ``[All, Web]``
            - If :attr:`type` is :attr:`.Type.Clan`:
                - :attr:`id` is non-zero
                - :attr:`instance` is ``All``.
            - If :attr:`type` is :attr:`.Type.GameServer`:
                - :attr:`id` is non-zero
        """
        if not (0 < self.id64 < 2**64):
            return False  # this shouldn't ever happen unless someone messes around with id64 but w/e
        if not (Universe.Invalid < self.universe <= Universe.Dev):
            return False
        if not (Type.Invalid < self.type <= Type.AnonUser):
            return False

        match self.type:
            case Type.Individual:
                return self.id != 0 and Instance.All <= self.instance <= Instance.Web
            case Type.Clan:
                return self.id != 0 and self.instance == Instance.All
            case Type.GameServer:
                return self.id != 0
        return True

    @staticmethod
    def from_id2(value: str, /) -> IndividualID | None:
        """Create an ID from a user's :attr:`id2`.

        Parameters
        ----------
        value
            The ID2 e.g. ``STEAM_1:0:1234``.

        Note
        ----
        The universe will be set to :attr:`Universe.Public` if it's ``0``. See :attr:`ID.id2_zero`.
        """
        if (match := ID2_REGEX.fullmatch(value)) is None:
            return None

        id = (int(match["id"]) << 1) | int(match["remainder"])
        universe = (
            int(match["universe"])
            or 1  # games before orange box used to incorrectly display universe as 0, we support that
        )

        return ID(id, type=Type.Individual, universe=Universe.try_value(universe), instance=Instance.Desktop)

    @staticmethod
    def from_id3(value: str, /) -> ID | None:
        """Create an ID from an SteamID's :attr:`id3`.

        Parameters
        ----------
        value
            The ID3 e.g. ``[U:1:1234]``.
        """
        if (match := ID3_REGEX.fullmatch(value)) is None:
            return None

        id = ID32(match["id"])
        universe = Universe.try_value(int(match["universe"]))
        type_char = TypeChar[match["type"].replace("i", "I")]
        if instance_ := match["instance"]:
            instance = Instance.try_value(int(instance_))
        else:
            instance = Instance.All
            # we can't use simple match because that uses the int.__eq__ which won't work for L, c and T
            match type_char.name:
                case TypeChar.g.name | TypeChar.T.name:
                    instance = Instance.All
                case TypeChar.L.name:
                    instance = Instance.ChatLobby
                case TypeChar.c.name:
                    instance = Instance.ChatClan
                case TypeChar.G.name | TypeChar.U.name:
                    instance = Instance.Desktop

        return ID(id, type=type_char.value, universe=universe, instance=instance)

    @staticmethod
    def from_invite_code(value: str, /) -> ID[Literal[Type.Individual]] | None:
        """Create an ID from a user's :attr:`invite_code`.

        Parameters
        ----------
        value
            The invite code e.g. ``cv-dgb``
        """
        if (search := INVITE_REGEX.fullmatch(value)) is None:
            return None

        code = search["code"].replace("-", "")

        id = ID32(_invite_custom_sub(code), 16)

        if 0 < id < 2**32:
            return ID(id, type=Type.Individual, universe=Universe.Public, instance=Instance.Desktop)

    @staticmethod
    async def from_url(
        url: StrOrURL, /, session: ClientSession | None = None
    ) -> ID[Literal[Type.Individual, Type.Clan]] | None:
        """Fetches the ID associated with a Steam Community URL or ``None``.

        Note
        -----
        Each call makes an HTTP request to https://steamcommunity.com.

        Examples
        --------
        The following are valid calls:

        .. code:: pycon

            >>> await ID.from_url("https://steamcommunity.com/groups/Valve")
            >>> await ID.from_url("https://steamcommunity.com/gid/103582791429521412")
            >>> await ID.from_url("https://steamcommunity.com/gid/[g:1:4]")
            ID(id=4, type=Type.Clan, universe=Universe.Public, instance=Instance.All)

            >>> await ID.from_url("https://steamcommunity.com/id/johnc")
            >>> await ID.from_url("https://steamcommunity.com/profiles/76561197960265740")
            >>> await ID.from_url("https://steamcommunity.com/profiles/[U:1:12]")
            >>> await ID.from_url("https://steamcommunity.com/user/r")
            ID(id=12, type=Type.Individual, universe=Universe.Public, instance=Instance.Desktop)

            >>> await ID.from_url("https://steamcommunity.com/app/570")
            ID(id=3703047, type=Type.Clan, universe=Universe.Public, instance=Instance.All)

        Parameters
        ----------
        url
            The Steam Community URL.
        session
            The session to make the request with. If ``None``, one is generated.
        """

        if not (search := URL_REGEX.match(str(url))):
            return None

        async with (
            aiohttp.ClientSession() if session is None else nullcontext(session) as session,
            session.get(f"https://{search['clean_url']}") as r,
        ):
            text = await r.text()

        if search["type"] in USER_URL_PATHS:
            data = JSON_LOADS(match["json"]) if (match := USER_ID64_FROM_URL_REGEX.search(text)) else None
        else:
            data = CLAN_ID64_FROM_URL_REGEX.search(text)
        return ID[Literal[Type.Individual, Type.Clan]](int(data["steamid"])) if data else None


ID_ZERO: Final = ID[Literal[Type.Individual]](0, type=Type.Individual)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import inspect
import logging
from contextvars import ContextVar
from types import CoroutineType
from typing import TYPE_CHECKING, Any, Final, Generic, TypeVar, cast, get_args

from .._const import CLEAR_PROTO_BIT, IS_PROTO
from ..app import App
from ..protobufs import GCMessage, GCProtobufMessage, friends
from ..protobufs.emsg import EMsg
from ..state import ConnectionState, ParserCallback
from ..trade import Inventory, Item
from ..types.id import AppID, AssetID, ContextID

if TYPE_CHECKING:
    from collections.abc import Mapping

    from typing_extensions import Self

    from ..gateway import GCMsgs
    from ..protobufs.client_server_2 import CMsgGcClientFromGC
    from ..utils import cached_property
    from .client import Client, ClientUser

log = logging.getLogger(__name__)
Inv = TypeVar("Inv", bound=Inventory[Item["ClientUser"], "ClientUser"])

APP = ContextVar[App]("APP")


class MultiEvent:
    def __init__(self, count: int):
        self.count = count
        self.ready = asyncio.Event()
        self._orig_count = self.count

    async def wait(self) -> None:
        await self.ready.wait()

    def set(self) -> None:
        self.count -= 1
        if self.count == 0:
            self.ready.set()

    def is_set(self) -> bool:
        return self.ready.is_set()

    def clear(self) -> None:
        self.count = self._orig_count
        self.ready.clear()


class GCState(ConnectionState, Generic[Inv]):
    gc_parsers: dict[
        type[GCMsgs], ParserCallback[Self, GCMsgs]
    ]  # different to parsers to save on dict lookups 1 vs 2 (1 for app, 1 for msg)
    client: Client
    _APP: Final[App]  # type: ignore
    if TYPE_CHECKING:

        @cached_property
        def user(self) -> ClientUser: ...

    def __init__(self, client: Client, **kwargs: Any):
        self._gc_connected = MultiEvent(len(client._GC_APPS))
        self._gc_ready = MultiEvent(len(client._GC_APPS))
        self.backpacks: Mapping[AppID, Inventory[Item[ClientUser], ClientUser]] = {}
        self.items_waiting: dict[tuple[AppID, AssetID], asyncio.Future[Item[ClientUser]]] = {}

        app = kwargs.pop("app", None)
        if app is not None:  # don't let them overwrite the main app
            try:
                kwargs["apps"].append(app)
            except (TypeError, KeyError):
                kwargs["apps"] = [app]
        kwargs["app"] = self._APP
        self._original_apps = [app for app in kwargs.get("apps", ()) if app.id not in self.client._GC_APPS]
        super().__init__(client, **kwargs)
        self._original_client_user_msg: friends.CMsgClientPersonaStateFriend | None = None

    def __init_subclass__(cls) -> None:
        cls.gc_parsers = {}
        for _, func in inspect.getmembers(cls, lambda x: inspect.isfunction(x) and hasattr(x, "__parser__")):
            try:
                params = list(inspect.get_annotations(func, eval_str=True).values())
            except NameError:
                continue
            msg = (
                args[0] if (args := get_args(params[0])) else params[0]
            )  # if it's a union only use the first type (Message | None or Message | Any)
            cls.gc_parsers[msg] = func

    @property
    def backpack(self) -> Inv | None:
        try:
            return cast(Inv, self.backpacks[APP.get().id])
        except KeyError:
            return None

    @backpack.setter
    def backpack(self, value: Inv) -> None:
        self.backpacks[APP.get().id] = value  # type: ignore

    def _get_gc_message(self) -> GCProtobufMessage | GCMessage | None:
        raise NotImplementedError()

    def parse_gc_message(self, msg: CMsgGcClientFromGC) -> None:
        app_id = AppID(msg.appid)
        emsg_value = CLEAR_PROTO_BIT(msg.msgtype)

        try:
            gc_msg = (GCProtobufMessage if IS_PROTO(msg.msgtype) else GCMessage)().parse(
                msg.payload[4:], emsg_value, app_id
            )
        except Exception as exc:
            return log.error("Failed to deserialize message: %r, %r", emsg_value, msg.payload, exc_info=exc)

        log.debug("Socket has received GC message %r from the websocket.", gc_msg)
        APP.set(self.client._GC_APPS[app_id])

        try:
            event_parser = self.gc_parsers[gc_msg.__class__]
        except (KeyError, TypeError):
            try:
                log.debug("Ignoring event %r", gc_msg, exc_info=True)
            except Exception:
                log.debug("Ignoring event with %r", gc_msg.__class__)
        else:
            try:
                result = event_parser(self, gc_msg)
            except Exception:
                return log.exception("Failed to execute %r", event_parser.__name__)

            if isinstance(result, CoroutineType):
                self.ws.tg.create_task(result, name=f"steam.py GC {app_id}: {event_parser.__name__}")

        # remove the dispatched listener
        removed: list[int] = []
        for idx, entry in enumerate(self.ws.gc_listeners):
            if entry.msg != emsg_value or entry.app_id != app_id:
                continue

            future = entry.future
            if future.cancelled():
                removed.append(idx)
                continue

            try:
                valid = entry.check(gc_msg)
            except Exception as exc:
                future.set_exception(exc)
                removed.append(idx)
            else:
                if valid:
                    future.set_result(gc_msg)
                    removed.append(idx)

        for idx in reversed(removed):
            del self.ws.gc_listeners[idx]

    async def fetch_backpack(self, backpack_cls: type[Inv]) -> Inv:
        app = APP.get()
        try:
            lock = self.user._inventory_locks[app.id]
        except KeyError:
            lock = self.user._inventory_locks[app.id] = asyncio.Lock()

        async with lock:  # requires a per-app lock to avoid Result.DuplicateRequest
            resp = await self.fetch_user_inventory(self.user.id64, app.id, ContextID(2), self.language)
        backpack = backpack_cls(
            state=self, proto=resp, owner=self.user, app=app, context_id=ContextID(2), language=self.language
        )
        self.backpacks[app] = backpack  # type: ignore
        return backpack

    def add_item_to_backpack(self, item: Item[ClientUser]) -> None:
        backpack = self.backpacks[item.app.id]
        backpack.items.append(item)  # type: ignore
        if future := self.items_waiting.get((item.app.id, item.id)):
            future.set_result(item)

    async def wait_for_item(self, id: AssetID) -> Item[ClientUser]:  # No intersection :( Item[ClientUser] & Any
        self.items_waiting[(APP.get().id, id)] = future = (
            asyncio.get_running_loop().create_future()
        )  # TODO does this need to use ContextID as well?
        return await future


GCState.parsers = ConnectionState.parsers | {EMsg.ClientFromGC: GCState.parse_gc_message}  # type: ignore



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from datetime import timedelta
from ipaddress import IPv4Address
from operator import attrgetter
from typing import TYPE_CHECKING, Annotated, Any, Literal, NamedTuple, cast, get_type_hints

from typing_extensions import TypedDict, Unpack

from .app import App, PartialApp
from .enums import GameServerRegion, Type
from .id import ID
from .protobufs.game_servers import EQueryType, GetServerListResponseServer, QueryResponse

if TYPE_CHECKING:
    from .state import ConnectionState


__all__ = (
    "Query",
    "GameServer",
    "ServerPlayer",
)


class Operator:
    @staticmethod
    def not_(arg: str) -> str:
        return Operator.nor(arg)

    @staticmethod
    def and_(*args: str) -> str:
        """All"""
        return "".join(args)

    # or would be nice if you can figure it out

    @staticmethod
    def nand(*args: str) -> str:
        """Not all"""
        joined = "\\".join(args)
        return f"\\nand\\{len(args)}{joined}"

    @staticmethod
    def nor(*args: str) -> str:
        """Not any"""
        joined = "\\".join(args)
        return f"\\nor\\{len(args)}{joined}"


# class LFD2SpecificQueries:  # TODO need implementing
#     @property
#     def match_hidden_tags(cls) -> Query[list[str]]:
#         """Fetches servers with all the given tag(s) in their 'hidden' tags only applies for :attr:`steam.LFD2`."""
#         return Query["list[str]"]("\\gamedata\\", type=list, callback=lambda items: f"[{','.join(items)}]")

#     @property
#     def match_any_hidden_tags(cls) -> Query[list[str]]:
#         """Fetches servers with any of the given tag(s) in their 'hidden' tags only applies for :attr:`steam.LFD2`."""
#         return Query["list[str]"]("\\gamedataor\\", type=list, callback=lambda items: f"[{','.join(items)}]")


class QueryWhereBoolKwargs(TypedDict, total=False):
    # Annotated[bool, TruthyCase | None, FalsyCase?]
    empty: Annotated[bool, r"\noplayers\1", r"\empty\1"]
    is_proxy: Annotated[bool, r"\proxy\1"]
    whitelisted: Annotated[bool, r"\white\1"]
    dedicated: Annotated[bool, r"\dedicated\1"]
    secure: Annotated[bool, r"\secure\1"]
    linux: Annotated[bool, r"\linux\1"]
    has_password: Annotated[bool, None, r"\password\0"]
    full: Annotated[bool, None, r"\full\1"]
    only_one_per_ip: Annotated[Literal[True], r"\collapse_addr_hash\1"]  # only makes sense as Literal[True]


class QueryWhereSimpleKwargs(TypedDict, total=False):
    version_match: Annotated[str, "\\version_match\\"]
    name_match: Annotated[str, "\\name_match\\"]
    game_directory: Annotated[str, "\\gamedir\\"]
    map: Annotated[str, "\\map\\"]
    ip: Annotated[str, "\\gameaddr\\"]
    app: Annotated[App, "\\app\\", attrgetter("id")]  # has inverse?
    not_app: Annotated[App, "\\napp\\", attrgetter("id")]
    tags: Annotated[list[str], "\\gametype\\", ",".join]
    region: Annotated[GameServerRegion, "\\region\\", attrgetter("value")]


class QueryWhereKwargs(QueryWhereBoolKwargs, QueryWhereSimpleKwargs, total=False): ...


for cls in (QueryWhereBoolKwargs, QueryWhereSimpleKwargs):
    cls.__annotations__ = get_type_hints(cls, include_extras=True)

QueryWhereKwargs.__annotations__ = QueryWhereBoolKwargs.__annotations__ | QueryWhereSimpleKwargs.__annotations__


class Query:
    """A class to construct Global Master Server queries."""

    @classmethod
    def where(cls, **kwargs: Unpack[QueryWhereKwargs]) -> str:
        """Construct a query meeting all the given criteria.

        Parameters
        ----------
        empty
            Fetches servers that are empty.
        is_proxy
            Fetches servers that are spectator proxies.
        whitelisted
            Fetches servers that are whitelisted.
        dedicated
            Fetches servers that are running dedicated.
        secure
            Fetches servers that are using anti-cheat technology (VAC, but potentially others as well).
        linux
            Fetches servers running on a Linux platform.
        has_password
            Fetches servers that are password protected.
        full
            Fetches servers that are full.
        only_one_per_ip
            Fetches only one server for each unique IP address matched.
        version_match
            Fetches servers running version "x" (``"*"`` is wildcard equivalent to ``.*`` in regex).
        name_match
            Fetches servers with their name matching "x" (``"*"`` is wildcard equivalent to ``.*`` in regex).
        game_directory
            Fetches servers running the specified modification (e.g. cstrike).
        map
            Fetches servers running the specified map (e.g. cs_italy)
        ip
            Fetches servers on the specified IP address, port is optional.
        app
            Fetches servers running a :class:`.App`.
        not_app
            Fetches servers not running a :class:`.App`.
        tags
            Fetches servers with all the given tag(s) in :attr:`GameServer.tags`.
        region
            Fetches servers in a given region.

        Returns
        -------
        The query that should be passed to :meth:`Client.fetch_servers`.

        Examples
        --------

        Match servers running TF2, that are not empty and are using VAC

        .. code-block:: pycon

            >>> Query.where(
            ...     app=TF2,
            ...     empty=False,
            ...     secure=True,
            ... )

        Match servers where the server name is not "A cool Server" and the server supports "alltalk" and increased max
        players

        .. code-block:: pycon

            >>> Query.where(name_match="A not cool server*", match_tags=["alltalk", "increased_maxplayers"])
        """

        filters: list[str] = []

        for name, value in kwargs.items():
            try:
                annotated = QueryWhereKwargs.__annotations__[name]
            except KeyError:
                raise TypeError(f"{name!r} is an invalid keyword argument for where()") from None
            metadata: tuple[Any, ...] = annotated.__metadata__
            if name in QueryWhereBoolKwargs.__annotations__:
                value = cast(bool, value)
                try:
                    filter_code = metadata[not value]
                except IndexError:
                    filter_code = None
                if filter_code is None:
                    inverse_filter_code = metadata[value]
                    filter_code = Operator.not_(inverse_filter_code)
            else:
                filter_code = metadata[0]
                if len(metadata) > 1:
                    value = str(metadata[-1](value))
                filter_code += value
            filters.append(filter_code)
        return Operator.and_(*filters)


class ServerPlayer(NamedTuple):
    name: str
    """The player's name."""
    score: int
    """The player's score."""
    play_time: timedelta
    """The amount of time the player has spent on the server."""


class GameServer(ID[Literal[Type.GameServer]]):
    """Represents a game server."""

    __slots__ = (
        "name",
        "app",
        "ip",
        "port",
        "tags",
        "map",
        "bot_count",
        "player_count",
        "max_player_count",
        "region",
        "version",
        "game_directory",
        "_secure",
        "_dedicated",
        "_state",
    )

    def __init__(self, state: ConnectionState, server: GetServerListResponseServer):
        super().__init__(server.steamid, type=Type.GameServer)
        self.name = server.name
        """The name of the server."""
        self.app = PartialApp(state, id=server.appid)
        """The app of the server."""
        self.ip = IPv4Address(server.addr.rpartition(":")[0])
        """The ip of the server."""
        self.port = server.gameport
        """The port of the server."""
        self.tags = server.gametype.split(",")
        """The tags of the server."""
        self.map = server.map
        """The map the server is running."""
        self.bot_count = server.bots
        """The number of bots in the server."""
        self.player_count = server.players
        """The number of players the server."""
        self.max_player_count = server.max_players
        """The maximum player count of the server."""
        self.region = GameServerRegion.try_value(server.region)
        """The region the server is in."""
        self.version = server.version
        """The version of the server."""
        self.game_directory = server.gamedir
        """The server's game directory e.g. ``"cstrike"``"""

        self._secure = server.secure
        self._dedicated = server.dedicated
        self._state = state

    def __repr__(self) -> str:
        attrs = ("name", "app", "ip", "port", "region", "id", "universe", "instance")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def __str__(self) -> str:
        return self.name

    def is_secure(self) -> bool:
        """Whether the sever is secured, likely with VAC."""
        return self._secure

    def is_dedicated(self) -> bool:
        """Whether the sever is dedicated."""
        return self._dedicated

    async def _query(self, type: EQueryType) -> QueryResponse:
        return await self._state.query_server(int(self.ip), self.port, self.app.id, type)

    async def update(self) -> None:
        """Update this game server instance in place."""
        proto = await self._query(EQueryType.Ping)
        server = proto.ping_data
        # self.ip = ip_address(betterproto.which_one_of(server, "server_ip")[1])
        # query_port: int = betterproto.uint32_field(2)
        # game_port: int = betterproto.uint32_field(3)
        # spectator_port: int = betterproto.uint32_field(4)
        # spectator_server_name: str = betterproto.string_field(5)
        self.name = server.server_name
        self.game_directory = server.gamedir
        self.map = server.map
        # game_description: str = betterproto.string_field(11)
        self.tags = server.gametype.split(",")
        self.player_count = server.num_players
        self.max_player_count = server.max_players
        self.bot_count = server.num_bots
        # password: bool = betterproto.bool_field(16)
        self._secure = server.secure
        self._dedicated = server.dedicated
        self.version = server.version
        # sdr_popid: int = betterproto.fixed32_field(20)
        # sdr_location_string: str = betterproto.string_field(21)

    async def players(self) -> list[ServerPlayer]:
        """Fetch a server's players.

        Returns
        -------
        .. source:: steam.ServerPlayer
        """
        proto = await self._query(EQueryType.Players)
        return [
            ServerPlayer(
                name=player.name,
                score=player.score,
                play_time=timedelta(seconds=player.time_played),
            )
            for player in proto.players_data.players
        ]

    async def rules(self) -> dict[str, str]:
        """Fetch a console variables. e.g. ``sv_gravity`` or ``sv_voiceenable``."""
        proto = await self._query(EQueryType.Rules)
        return proto.rules_data.rules



from ....protobufs.msg import GCMessage
from ....utils import StructIO
from ..enums import EMsg


class NameItemRequest(GCMessage, msg=EMsg.NameItem):
    name_tag_id: int
    item_id: int
    name: str

    def __bytes__(self) -> bytes:
        with StructIO() as io:
            io.write_u64(self.name_tag_id)
            io.write_u64(self.item_id)
            io.write(b"\x00" + self.name.encode("UTF-8") + b"\x00")
            return io.buffer


class DeleteItemRequest(GCMessage, msg=EMsg.Delete):
    item_id: int



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import asyncio
import contextlib
import itertools
import types
from collections.abc import Iterator, Sequence
from typing import TYPE_CHECKING, Any, Generic, cast, overload

from typing_extensions import NamedTuple, TypeVar, get_original_bases

from . import utils
from ._const import URL
from .app import App, PartialApp
from .enums import Language, TradeOfferState
from .models import AssetMixin, DescriptionMixin
from .protobufs import econ
from .types.id import AppID, AssetID, ClassID, ContextID, InstanceID, TradeOfferID
from .utils import DateTime

if TYPE_CHECKING:
    from datetime import datetime, timedelta

    from .abc import BaseUser, PartialUser
    from .friend import Friend
    from .state import ConnectionState
    from .types import trade
    from .user import ClientUser, User


__all__ = (
    "Asset",
    "Item",
    "Inventory",
    "TradeOffer",
    "MovedItem",
    "TradeOfferReceipt",
)


OwnerT = TypeVar("OwnerT", bound="PartialUser", default="BaseUser", covariant=True)


class Asset(AssetMixin, Generic[OwnerT]):
    """Base most version of an item. This class should only be received when Steam fails to find a matching item for
    its class and instance IDs.

    .. container:: operations

        .. describe:: x == y

            Checks if two assets are equal.

        .. describe:: hash(x)

            Returns the hash of an asset.
    """

    __slots__ = (
        "id",
        "amount",
        "instance_id",
        "context_id",
        # "post_rollback_id",
        "owner",
    )
    REPR_ATTRS = ("id", "class_id", "instance_id", "context_id", "amount", "owner", "app")  # "post_rollback_id"

    def __init__(self, state: ConnectionState, asset: econ.Asset, owner: OwnerT):
        self.id = AssetID(asset.assetid)
        """The assetid of the item."""
        self.amount = asset.amount
        """The amount of the same asset there are in the inventory."""
        self.instance_id = InstanceID(asset.instanceid)
        """The instanceid of the item."""
        self.class_id = ClassID(asset.classid)
        self.context_id = ContextID(asset.contextid)
        """The contextid of the item."""
        # self.post_rollback_id = int(data["rollback_new_assetid"]) if "rollback_new_assetid" in data else None
        """The assetid of the item after a rollback (cancelled, etc.). ``None`` if not rolled back."""
        self.owner = owner
        """The owner of the asset."""
        self._app_id = AppID(asset.appid)
        self._state = state

    def __repr__(self) -> str:
        cls = self.__class__
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in cls.REPR_ATTRS]
        return f"<{cls.__name__} {' '.join(resolved)}>"

    def __eq__(self, other: object) -> bool:
        return (
            isinstance(other, Asset)
            and self.id == other.id
            and self._app_id == other._app_id
            and self.context_id == other.context_id
        )

    def __hash__(self) -> int:
        return hash((self.id, self._app_id, self.context_id))

    def to_dict(self) -> trade.AssetToDict:
        return {
            "assetid": str(self.id),
            "amount": self.amount,
            "appid": str(self._app_id),
            "contextid": str(self.context_id),
        }

    def to_proto(self) -> econ.Asset:
        return econ.Asset(
            assetid=self.id,
            amount=self.amount,
            instanceid=self.instance_id,
            classid=self.class_id,
            appid=self._app_id,
            contextid=self.context_id,
        )

    @property
    def url(self) -> str:
        """The URL for the asset in the owner's inventory.

        e.g. https://steamcommunity.com/profiles/76561198248053954/inventory/#440_2_8526584188
        """
        return f"{self.owner.community_url}/inventory#{self._app_id}_{self.context_id}_{self.id}"

    async def gift_to(
        self: Asset[ClientUser],
        recipient: Friend,
        /,
        *,
        name: str | None = None,
        message: str,
        closing_note: str,
        signature: str,
    ) -> None:
        """Gifts the asset to the recipient.

        Parameters
        ------------
        recipient
            The recipient to gift the asset to.
        name
            The name to give the asset. If not provided, the recipient's :attr:`name` will be used.
        message
            The message to send with the gift.
        closing_note
            The closing note to send with the gift.
        signature
            The signature to send with the gift.
        """
        await self._state.http.send_user_gift(
            recipient.id,
            self.id,
            name=name if name is not None else recipient.name,
            message=message,
            closing_note=closing_note,
            signature=signature,
        )


class Item(Asset[OwnerT], DescriptionMixin):
    """Represents an item in a User's inventory."""

    __slots__ = DescriptionMixin.SLOTS
    REPR_ATTRS = ("name", *Asset.REPR_ATTRS)

    def __init__(self, state: ConnectionState, asset: econ.Asset, description: econ.ItemDescription, owner: OwnerT):
        super().__init__(state, asset, owner)
        DescriptionMixin.__init__(self, state, description)


class InventoryGenericAlias(types.GenericAlias):
    if TYPE_CHECKING:

        @property
        def __origin__(self) -> type[Inventory]: ...

    def __call__(self, *args: Any, **kwargs: Any) -> object:
        # this is done cause we need __orig_class__ in __init__
        result = self.__origin__.__new__(self.__origin__, *args, **kwargs)
        result.__orig_class__ = self
        result.__init__(*args, **kwargs)
        return result

    def __mro_entries__(self, bases: tuple[type, ...]) -> tuple[type]:
        # if we are subclassing we should return a new class that already has __orig_class__

        class BaseInventory(*super().__mro_entries__(bases)):  # type: ignore
            __slots__ = ()
            __orig_class__ = self

        return (BaseInventory,)


ItemT = TypeVar("ItemT", bound=Item["PartialUser"], default=Item["BaseUser"], covariant=True)


class Inventory(Generic[ItemT, OwnerT], Sequence[ItemT]):
    """Represents a User's inventory.

    .. container:: operations

        .. describe:: len(x)

            Returns how many items are in the inventory.

        .. describe:: iter(x)

            Iterates over the inventory's items.

        .. describe:: x[i]

            Returns the item at the given index.

        .. describe:: y in x

            Determines if an item is in the inventory based off of its :attr:`Asset.id`.
    """

    __slots__ = (
        "app",
        "items",
        "owner",
        "context_id",
        "_language",
        "_state",
        "__orig_class__",  # undocumented typing internals more shim to make extensions work
    )
    __orig_class__: InventoryGenericAlias

    def __init__(
        self,
        state: ConnectionState,
        proto: econ.GetInventoryItemsWithDescriptionsResponse,
        owner: OwnerT,
        app: App,
        context_id: ContextID,
        language: Language | None,
    ):
        self._state = state
        self.owner = owner
        """The owner of the inventory."""
        self.app = PartialApp(state, id=app.id, name=app.name)
        """The app the inventory belongs to."""
        self.context_id = context_id
        self._language = language
        self._update(proto)

    def __repr__(self) -> str:
        attrs = ("owner", "app")
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in attrs]
        try:
            name = type(self).__name__
        except AttributeError:
            name = self.__orig_class__.__name__
        return f"<{name} {' '.join(resolved)}>"

    def __len__(self) -> int:
        return len(self.items)

    if not TYPE_CHECKING:
        __class_getitem__ = classmethod(InventoryGenericAlias)

        def __getitem__(self, idx):
            return self.items[idx]

    def __iter__(self) -> Iterator[ItemT]:
        return iter(self.items)

    def __contains__(self, item: object) -> bool:
        return item in self.items

    def _update(self, proto: econ.GetInventoryItemsWithDescriptionsResponse, /) -> None:
        items: list[ItemT] = []
        ItemClass: type[ItemT]
        try:  # ideally one day this will just be ItemT.__value__ or something
            ItemClass, *_ = self.__orig_class__.__args__
        except AttributeError:
            ItemClass = get_original_bases(self.__class__)[0].__args__[0].__default__
        for asset in proto.assets:
            description = utils.get(proto.descriptions, instanceid=asset.instanceid, classid=asset.classid)
            if description is None:
                raise RuntimeError(f"Associated description for {asset} not found")
            items.append(ItemClass(self._state, asset=asset, description=description, owner=self.owner))
        self.items: Sequence[ItemT] = items
        """A list of the inventory's items."""

    async def update(self) -> None:
        """Re-fetches the inventory and updates it inplace."""
        async with (
            self._state.user._inventory_locks.setdefault(self.app.id, asyncio.Lock())
            if self.owner == self._state.user
            else contextlib.nullcontext()
        ):
            proto = await self._state.fetch_user_inventory(
                self.owner.id64, self.app.id, self.context_id, self._language
            )
        self._update(proto)


class TradeOfferReceipt(NamedTuple, Generic[OwnerT]):
    sent: list[MovedItem[ClientUser]]
    received: list[MovedItem[OwnerT]]


class MovedItem(Item[OwnerT]):
    """Represents an item that has moved from one inventory to another."""

    __slots__ = (
        "new_id",
        "new_context_id",
    )
    REPR_ATTRS = (*Item.REPR_ATTRS, "new_id", "new_context_id")

    def __init__(self, state: ConnectionState, data: trade.TradeOfferReceiptItem, owner: OwnerT):
        super().__init__(
            state,
            asset=econ.Asset().from_dict(data),
            description=econ.ItemDescription().from_dict(data),
            owner=owner,
        )
        try:
            self.new_id = AssetID(
                int(data["new_assetid"]) if "new_assetid" in data else int(data["rollback_new_assetid"])
            )
            """The new_assetid field, this is the asset ID of the item in the partners inventory."""
        except KeyError:
            self.new_id = AssetID(-1)  # steam broke?

        try:
            self.new_context_id = ContextID(
                int(data["new_contextid"]) if "new_contextid" in data else int(data["rollback_new_contextid"])
            )
            """The new_contextid field."""
        except KeyError:
            self.new_context_id = ContextID(-1)  # steam broke again probably


ReceivingAssetT = TypeVar("ReceivingAssetT", bound="Asset[PartialUser]", default="Item[User]", covariant=True)
SendingAssetT = TypeVar("SendingAssetT", bound="Asset[ClientUser]", default="Item[ClientUser]", covariant=True)
UserT = TypeVar("UserT", bound="PartialUser", default="User", covariant=True)


class TradeOffer(Generic[ReceivingAssetT, SendingAssetT, UserT]):
    """Represents a trade offer from/to send to a User.
    This can also be used in :meth:`steam.User.send`.

    Parameters
    ----------
    sending
        The items you are sending to the other user.
    receiving
        The items you are receiving from the other user.
    token
        The trade token used to send trades to users who aren't on the ClientUser's friend's list.
    message
         The offer message to send with the trade.
    """

    __slots__ = (
        "id",
        "_id",
        "state",
        "escrow",
        "user",
        "message",
        "token",
        "expires",
        "updated_at",
        "created_at",
        "sending",
        "receiving",
        "_has_been_sent",
        "_state",
        "_is_our_offer",
    )

    id: TradeOfferID
    """The trade offer's ID."""
    user: UserT
    """The trade offer partner."""

    def __init__(
        self,
        *,
        message: str | None = None,
        token: str | None = None,
        sending: Sequence[SendingAssetT] | SendingAssetT | None = None,
        # TODO HKT for this would be really nice as could then "ensure" we own the item
        receiving: Sequence[ReceivingAssetT] | ReceivingAssetT | None = None,
    ):
        self.sending: Sequence[SendingAssetT] = (
            sending if isinstance(sending, Sequence) else [sending] if sending else []
        )
        """The items you are sending to the partner."""
        self.receiving: Sequence[ReceivingAssetT] = (
            receiving if isinstance(receiving, Sequence) else [receiving] if receiving else []
        )
        """The items you are receiving from the partner."""
        self.message: str | None = message or None
        """The message included with the trade offer."""
        self.token: str | None = token
        """The trade token used to send trades to users who aren't on the ClientUser's friend's list."""
        self.updated_at: datetime | None = None
        """The time at which the trade was last updated."""
        self.created_at: datetime | None = None
        """The time at which the trade was created."""
        self.escrow: timedelta | None = None
        """
        The time at which the escrow (hold) will end. Can be ``None`` if there is no escrow on the trade.

        Warning
        -------
        This isn't likely to be accurate, use :meth:`User.escrow` instead if possible.
        """
        self.state = TradeOfferState.Invalid
        """The offer state of the trade for the possible types see :class:`~steam.TradeOfferState`."""
        self._id: int | None = None
        self._has_been_sent = False

    @classmethod
    def _from_api(
        cls,
        state: ConnectionState,
        data: trade.TradeOffer,
        sending: list[tuple[econ.Asset, econ.ItemDescription]],
        receiving: list[tuple[econ.Asset, econ.ItemDescription]],
        user: UserT,  # type: ignore
    ) -> TradeOffer[Item[UserT], Item[ClientUser], UserT]:
        trade = cls()
        trade._has_been_sent = True
        trade._state = state
        trade.user = user
        return cast(
            "TradeOffer[Item[UserT], Item[ClientUser], UserT]",
            trade._update(data, sending=sending, receiving=receiving),
        )

    @classmethod
    def _from_history(
        cls: type[TradeOffer[MovedItem[UserT], MovedItem[ClientUser], UserT]],
        state: ConnectionState,
        data: trade.TradeOfferHistoryTrade,
        descriptions: Sequence[trade.Description],
    ) -> TradeOffer[MovedItem[UserT], MovedItem[ClientUser], UserT]:
        user = cast("UserT", state.get_partial_user(data["steamid_other"]))
        trade = cls(
            receiving=[
                MovedItem(state, description | asset, user)
                for asset, description in itertools.product(data.get("assets_received", ()), descriptions)
                if description["instanceid"] == asset["instanceid"] and description["classid"] == asset["classid"]
            ],
            sending=[
                MovedItem(state, description | asset, state.user)
                for asset, description in itertools.product(data.get("assets_given", ()), descriptions)
                if description["instanceid"] == asset["instanceid"] and description["classid"] == asset["classid"]
            ],
        )
        trade._state = state
        trade._id = int(data["tradeid"])
        trade.user = user
        trade.created_at = DateTime.from_timestamp(data["time_init"])
        trade.state = TradeOfferState.try_value(data["status"])

        return trade

    def _update_from_send(
        self, state: ConnectionState, data: trade.TradeOfferCreateResponse, user: UserT, active: bool = True  # type: ignore
    ) -> None:
        self.id = TradeOfferID(int(data["tradeofferid"]))
        self._state = state
        self.user = user
        self.state = TradeOfferState.Active if active else TradeOfferState.ConfirmationNeed
        self.created_at = DateTime.now()
        self._is_our_offer = True

    def __repr__(self) -> str:
        attrs = ("id", "state", "user")
        resolved = [f"{attr}={getattr(self, attr, None)!r}" for attr in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    @overload
    def _update(
        self: TradeOffer[Asset[UserT], Asset[ClientUser], UserT],
        data: trade.TradeOffer,
        sending: list[tuple[econ.Asset, econ.ItemDescription]],
        receiving: list[tuple[econ.Asset, econ.ItemDescription]],
    ) -> TradeOffer[Item[UserT], Item[ClientUser], UserT]: ...

    @overload
    def _update(
        self: TradeOffer[Asset[UserT], Asset[ClientUser], UserT],
        data: trade.TradeOffer,
        sending: list[econ.Asset],
        receiving: list[econ.Asset],
    ) -> TradeOffer[Asset[UserT], Asset[ClientUser], UserT]: ...

    def _update(
        self: TradeOffer[Asset[UserT], Asset[ClientUser], UserT],
        data: trade.TradeOffer,
        sending: list[tuple[econ.Asset, econ.ItemDescription]] | list[econ.Asset],
        receiving: list[tuple[econ.Asset, econ.ItemDescription]] | list[econ.Asset],
    ) -> TradeOffer[Item[UserT], Item[ClientUser], UserT] | TradeOffer[Asset[UserT], Asset[ClientUser], UserT]:
        self.message = data.get("message") or None
        self.id = TradeOfferID(int(data["tradeofferid"]))
        self._id = int(data["tradeid"]) if "tradeid" in data else None
        expires = data.get("expiration_time")
        escrow = data.get("escrow_end_date")
        updated_at = data.get("time_updated")
        created_at = data.get("time_created")
        self.expires = DateTime.from_timestamp(expires) if expires else None
        self.escrow = DateTime.from_timestamp(escrow) - DateTime.now() if escrow else None
        self.updated_at = DateTime.from_timestamp(updated_at) if updated_at else None
        self.created_at = DateTime.from_timestamp(created_at) if created_at else None
        self.state = TradeOfferState.try_value(data.get("trade_offer_state", 1))
        self._is_our_offer = data.get("is_our_offer", False)
        if (self.state != TradeOfferState.Accepted) and (
            sending or receiving
        ):  # steam doesn't really send the item data if the offer just got accepted
            try:
                self.sending = [
                    Item(self._state, asset=asset, description=description, owner=self._state.user)
                    for asset, description in cast("list[tuple[econ.Asset, econ.ItemDescription]]", sending)
                ]
                self.receiving = [
                    Item(self._state, asset=asset, description=description, owner=self.user)
                    for asset, description in cast("list[tuple[econ.Asset, econ.ItemDescription]]", receiving)
                ]
                return cast("TradeOffer[Item[UserT], Item[ClientUser], UserT]", self)
            except TypeError:
                self.sending = [
                    Asset(self._state, asset=asset, owner=self._state.user)
                    for asset in cast("list[econ.Asset]", sending)
                ]
                self.receiving = [
                    Asset(self._state, asset=asset, owner=self.user) for asset in cast("list[econ.Asset]", receiving)
                ]
        return self

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, TradeOffer):
            return False
        return (
            self.id == other.id
            if self._has_been_sent and other._has_been_sent
            else self.sending == other.sending and self.receiving == other.receiving
        )

    def __hash__(self) -> int:
        return hash(self.id) if self._has_been_sent else hash((tuple(self.sending), tuple(self.receiving)))

    async def confirm(self) -> None:
        """Confirms the trade offer.
        This rarely needs to be called as the client handles most of these.

        Raises
        ------
        ValueError
            The trade is not active.
        steam.ConfirmationError
            No matching confirmation could not be found.
        """
        self._check_active()
        if self.is_gift():
            return  # no point trying to confirm it
        confirmation = await self._state.wait_for_confirmation(self.id)
        await confirmation.confirm()
        self._state._confirmations.pop(self.id, None)

    async def accept(self) -> None:
        """Accepts the trade offer.

        Note
        ----
        This also calls :meth:`confirm` (if necessary) so you don't have to.

        Raises
        ------
        ValueError
            The trade is either not active, already accepted or not from the ClientUser.
        ConfirmationError
            No matching confirmation could not be found.
        """
        if self.state == TradeOfferState.Accepted:
            raise ValueError("This trade has already been accepted")
        if self.is_our_offer():
            raise ValueError("You cannot accept an offer the ClientUser has made")
        self._check_active()
        assert self.user is not None
        resp = await self._state.http.accept_user_trade(self.user.id64, self.id)
        if resp.get("needs_mobile_confirmation", False):
            await self.confirm()

    async def decline(self) -> None:
        """Declines the trade offer.

        Raises
        ------
        ValueError
            The trade is either not active, already declined or not from the ClientUser.
        """
        if self.state == TradeOfferState.Declined:
            raise ValueError("This trade has already been declined")
        if self.is_our_offer():
            raise ValueError("You cannot decline an offer the ClientUser has made")
        self._check_active()
        await self._state.http.decline_user_trade(self.id)

    async def cancel(self) -> None:
        """Cancels the trade offer.

        Raises
        ------
        ValueError
            The trade is either not active or already cancelled.
        """
        if self.state == TradeOfferState.Canceled:
            raise ValueError("This trade has already been cancelled")
        self._check_active()
        await self._state.http.cancel_user_trade(self.id)

    async def receipt(self) -> TradeOfferReceipt[UserT]:
        """Get the receipt for a trade offer and the updated asset ids for the trade.

        .. source:: steam.TradeOfferReceipt

        Returns
        -------
        A trade receipt.
        """
        if self._id is None:
            raise ValueError("Cannot fetch the receipt for a trade not accepted")

        data = await self._state.http.get_trade_receipt(self._id)
        (trade,) = data["trades"]
        descriptions = data["descriptions"]
        assert self.user is not None

        return TradeOfferReceipt(
            sent=[
                MovedItem(self._state, data={**description, **asset}, owner=self._state.user)
                for asset, description in itertools.product(trade.get("assets_given", ()), descriptions)
                if description["instanceid"] == asset["instanceid"] and description["classid"] == asset["classid"]
            ],
            received=[
                MovedItem(self._state, data={**description, **asset}, owner=self.user)
                for asset, description in itertools.product(trade.get("assets_received", ()), descriptions)
                if description["instanceid"] == asset["instanceid"] and description["classid"] == asset["classid"]
            ],
        )

    async def counter(
        self: TradeOffer[Asset[UserT], Asset[ClientUser], UserT],
        trade: TradeOffer[Asset[UserT], Asset[ClientUser], Any],
        /,
    ) -> None:
        """Counter a trade offer from an :class:`User`.

        Parameters
        -----------
        trade
            The trade offer to counter with.

        Raises
        ------
        ValueError
            The trade from the ClientUser or it isn't active.
        """
        self._check_active()
        if self.is_our_offer():
            raise ValueError("You cannot counter an offer the ClientUser has made")

        assert self.user is not None
        await self.user._send_trade(trade, tradeofferid_countered=self.id)

    @property
    def url(self) -> str:
        """The URL of the trade offer."""
        return str(URL.COMMUNITY / f"tradeoffer/{self.id}")

    def is_gift(self) -> bool:
        """Helper method that checks if an offer is a gift to the :class:`~steam.ClientUser`"""
        return bool(self.receiving and not self.sending)

    def is_our_offer(self) -> bool:
        """Whether the offer was created by the :class:`~steam.ClientUser`."""
        return self._is_our_offer

    def _check_active(self) -> None:
        if self.state not in (TradeOfferState.Active, TradeOfferState.ConfirmationNeed) or not self._has_been_sent:
            raise ValueError("This trade is not active")



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Literal, TypedDict, final

from .vdf import TypedVDFDict, VDFBool, VDFInt, VDFList

if TYPE_CHECKING:
    from multidict import MultiDict
    from typing_extensions import NotRequired


class UserAppStats(TypedVDFDict):
    gamename: str
    version: VDFInt
    stats: VDFList[AppStat | AchievementStat]


class AppStatDisplay(TypedVDFDict):
    name: str


@final
class AppStat(TypedVDFDict):
    type: Literal["1"]
    name: str
    incrementonly: VDFBool
    display: AppStatDisplay
    id: VDFInt


class AchievementStatBitsDisplay(TypedVDFDict):
    name: MultiDict[str]
    """e.g.
    {
        "token": "TF_PLAY_GAME_FRIENDSONLY_NAME",
        "english": "With Friends Like these...",
        "german": "Kumpels",
    }
    """
    desc: MultiDict[str]  # same as above
    hidden: VDFBool
    icon: str  # https://cdn.akamai.steamstatic.com/steamcommunity/public/images/apps/440/ + this
    icon_gray: str


class AchievementStatBitsProgress(TypedVDFDict):
    min_value: NotRequired[VDFInt]
    max_value: NotRequired[VDFInt]
    value: MultiDict[str]


class AchievementStatBits(TypedVDFDict):
    name: str
    display: AchievementStatBitsDisplay
    progress: NotRequired[AchievementStatBitsProgress]
    bit: int  # actually an int lol


@final
class AchievementStat(TypedVDFDict):
    type: Literal["4"]
    bits: VDFList[AchievementStatBits]
    id: VDFInt


class AppAppStats(TypedDict):
    gameName: str
    gameVersion: int
    availableGameStats: AppAppStatsAvailableGameStats


class AppAppStatsAvailableGameStats(TypedDict):
    stats: NotRequired[list[AppAppStatsAvailableGameStatsStats]]
    achievements: NotRequired[list[AppStatsAvailableGameStatsAchievements]]


class AppAppStatsAvailableGameStatsStats(TypedDict):
    name: str
    defaultvalue: int
    displayName: str


class AppStatsAvailableGameStatsAchievements(TypedDict):
    name: str
    defaultvalue: int
    displayName: str
    hidden: bool  # 0 or 1
    description: str
    icon: str
    icongray: str



# pyright: reportIncompatibleMethodOverride = none
"""A nice way to work with TF2's currencies."""

from __future__ import annotations

import math
from decimal import Decimal
from fractions import Fraction
from functools import reduce
from typing import TYPE_CHECKING, TypeAlias, overload

if TYPE_CHECKING:
    from collections.abc import Iterable

    from typing_extensions import Self

    from ...trade import Item


__all__ = ("Metal",)


SupportsMetal: TypeAlias = int | str | float | Fraction | Decimal


def modf(value: Decimal) -> tuple[Decimal, Decimal]:
    # can't just use divmod(value, 1) as both raise for large Decimals because lol,
    # so we need this whole song and dance
    as_tuple = value.as_tuple()
    if not isinstance(as_tuple.exponent, int):
        raise ValueError(f"invalid exponent {as_tuple.exponent!r}")

    if not as_tuple.exponent:
        integer = value
        fractional = Decimal().copy_sign(value)
    elif as_tuple.exponent < 0:
        integer = Decimal((as_tuple.sign, as_tuple.digits[: as_tuple.exponent], 0))
        fractional = Decimal((as_tuple.sign, as_tuple.digits[as_tuple.exponent :], as_tuple.exponent))
    else:
        raise AssertionError("shouldn't be reachable")  # you get rounding issues here either way

    return fractional, integer


# TODO use https://github.com/python/cpython/pull/106223 when it's merged
class Metal(Fraction):
    """A class to represent some metal in TF2.

    The value used as the denominator corresponds to one scrap metal.

    Examples
    --------

    .. code:: python

        Metal("1.22") == Metal(1.22)
        Metal(1.22) == Metal(1.22)
        Metal(1) == Metal(1.00)

    Note
    ----
    When working with large floating point numbers, it's recommended to pass a :class:`str` to avoid precision loss.
    """

    __slots__ = ()

    @overload
    def __new__(cls, value: SupportsMetal, /) -> Self:  # type: ignore
        ...

    def __new__(cls, value: SupportsMetal, /, *, _normalize: bool = True) -> Self:
        return super().__new__(cls, cls.extract_scrap(value), 9)

    @classmethod
    def extract_scrap(cls, value: SupportsMetal) -> int:
        if isinstance(value, Fraction):
            if value.denominator in {9, 3, 1}:
                return value.numerator * (9 // value.denominator)
            raise ValueError("cannot convert Fraction to Metal, denominator isn't 1, 3 or 9")

        try:
            value = Decimal(value)
        except (ValueError, TypeError):
            raise TypeError("non-int passed to Metal.__new__, that could not be cast") from None

        fractional, integer = modf(value)
        rounded_fractional = round(fractional, 2)
        if not math.isclose(fractional - rounded_fractional, 0, abs_tol=1e-9):
            raise ValueError("metal value's last digits must be close to 0")

        digits = rounded_fractional.as_tuple().digits
        if len(digits) >= 2 and digits[0] != digits[1]:
            raise ValueError("metal value must be a multiple of 0.11")

        return int(integer) * 9 + (digits[0] if fractional >= 0 else -digits[0])

    def __add__(self, other: SupportsMetal) -> Metal:
        return Metal(super().__add__(Fraction(self.extract_scrap(other), 9)))

    def __sub__(self, other: SupportsMetal) -> Metal:
        return Metal(super().__sub__(Fraction(self.extract_scrap(other), 9)))

    def __mul__(self, other: SupportsMetal) -> Metal:
        return Metal(super().__mul__(Fraction(self.extract_scrap(other), 9)))

    def __truediv__(self, other: SupportsMetal) -> Metal:
        return Metal(super().__truediv__(Fraction(self.extract_scrap(other), 9)))

    __radd__ = __add__  # type: ignore
    __rsub__ = __sub__  # type: ignore
    __rmul__ = __mul__  # type: ignore
    __rtruediv__ = __truediv__  # type: ignore

    def __abs__(self) -> Metal:
        return Metal(super().__abs__())

    def __pos__(self) -> Metal:
        return Metal(super().__pos__())

    def __neg__(self) -> Metal:
        return Metal(super().__neg__())

    def __str__(self) -> str:
        return (
            f"{self.numerator // self.denominator}"
            "."
            f"{f'{(self.numerator % self.denominator) * 9 // self.denominator}' * 2}"
        )

    def __repr__(self) -> str:
        return f"{self.__class__.__name__}({str(self)!r})"

    @classmethod
    def from_items(cls, items: Iterable[Item]) -> Metal:
        """The amount of metal in the items.

        Example
        -------

        .. code:: py

            their_inventory = await user.inventory(steam.TF2)
            their_metal = Metal.from_items(their_inventory)
            our_inventory = await client.user.inventory(steam.TF2)
            await user.send(f"I have {'more' if our_inventory.metal > their_metal else 'less'} metal than you")
        """
        return reduce(cls.__add__, (ITEM_VALUES[item.name] for item in items if item.name in ITEM_VALUES))


ITEM_VALUES = {
    "Refined Metal": Metal(1.00),
    "Reclaimed Metal": Metal(0.33),
    "Scrap Metal": Metal(0.11),
}



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import abc
import re
from contextlib import AbstractAsyncContextManager, asynccontextmanager
from dataclasses import dataclass, field
from datetime import timedelta
from io import BytesIO
from typing import TYPE_CHECKING, Any, Literal, ParamSpec, TypedDict, TypeVar, runtime_checkable

from typing_extensions import Protocol

from ._const import DEFAULT_AVATAR, URL, ReadOnly
from .enums import Currency, PurchaseResult, Realm, Result
from .media import Media
from .types.id import AppID, ClassID

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Callable

    from _typeshed import StrOrBytesPath
    from aiohttp.streams import AsyncStreamIterator, ChunkTupleAsyncStreamIterator
    from yarl import URL as _URL

    from .app import PartialApp
    from .protobufs import econ
    from .state import ConnectionState
    from .types import user


__all__ = (
    "PriceOverview",
    "Ban",
    "Avatar",
    "Wallet",
)

F = TypeVar("F", bound="Callable[..., Any]")
P = ParamSpec("P")


def api_route(path: str, version: int = 1, *, publisher: bool = False) -> _URL:
    return (URL.PUBLISHER_API if publisher else URL.API) / f"{path}/v{version}"


class _ReturnTrue:
    __slots__ = ()

    def __call__(self, *_: Any, **__: Any) -> Literal[True]:
        return True

    def __repr__(self) -> str:
        return "<return_true>"


return_true = _ReturnTrue()


PRICE_RE = re.compile(r"(^\D*(?P<price>[\d,.]*)\D*$)")


class PriceOverviewDict(TypedDict):
    success: bool
    lowest_price: str
    median_price: str
    volume: str


class PriceOverview:
    """Represents the data received from the Steam Community Market."""

    __slots__ = ("currency", "volume", "lowest_price", "median_price")

    lowest_price: float | str
    """The lowest price observed by the market."""
    median_price: float | str
    """The median price observed by the market."""

    def __init__(self, data: PriceOverviewDict, currency: Currency) -> None:
        lowest_price_ = PRICE_RE.search(data["lowest_price"])
        median_price_ = PRICE_RE.search(data["median_price"])
        assert lowest_price_ is not None
        assert median_price_ is not None
        lowest_price = lowest_price_["price"]
        median_price = median_price_["price"]

        try:
            self.lowest_price = float(lowest_price.replace(",", "."))
            self.median_price = float(median_price.replace(",", "."))
        except ValueError:
            self.lowest_price = lowest_price
            self.median_price = median_price

        self.volume: int = int(data["volume"].replace(",", ""))
        """The number of items sold in the last 24 hours."""
        self.currency = currency

    def __repr__(self) -> str:
        resolved = [f"{attr}={getattr(self, attr)!r}" for attr in self.__slots__]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"


class Ban:
    """Represents a Steam ban."""

    __slots__ = (
        "since_last_ban",
        "number_of_game_bans",
        "_vac_banned",
        "_community_banned",
        "_market_banned",
    )

    def __init__(self, data: user.UserBan):
        self._vac_banned = data["vac_banned"]
        self._community_banned = data["community_banned"]
        self._market_banned = data["economy_ban"] != "none"
        self.since_last_ban = timedelta(days=data["days_since_last_ban"])
        """The amount of time that has passed since user was last banned"""
        self.number_of_game_bans: int = data["number_of_game_bans"]
        """The number of game bans the user has."""

    def __repr__(self) -> str:
        attrs = [
            ("number_of_game_bans", self.number_of_game_bans),
            ("is_vac_banned()", self.is_vac_banned()),
            ("is_community_banned()", self.is_community_banned()),
            ("is_market_banned()", self.is_market_banned()),
        ]
        resolved = [f"{method}={value!r}" for method, value in attrs]
        return f"<{self.__class__.__name__} {' '.join(resolved)}>"

    def is_banned(self) -> bool:
        """Species if the user is banned from any part of Steam."""
        return any((self.is_vac_banned(), self.is_community_banned(), self.is_market_banned()))

    def is_vac_banned(self) -> bool:
        """Whether or not the user is VAC banned."""
        return self._vac_banned

    def is_community_banned(self) -> bool:
        """Whether or not the user is community banned."""
        return self._community_banned

    def is_market_banned(self) -> bool:
        """Whether or not the user is market banned."""
        return self._market_banned


class StreamReaderProto(Protocol):
    def __aiter__(self) -> AsyncStreamIterator[bytes]: ...

    def iter_chunked(self, n: int) -> AsyncStreamIterator[bytes]: ...

    def iter_any(self) -> AsyncStreamIterator[bytes]: ...

    def iter_chunks(self) -> ChunkTupleAsyncStreamIterator: ...

    async def readline(self) -> bytes: ...

    async def readuntil(self, separator: bytes = b"\n", /) -> bytes: ...

    async def read(self, n: int = -1, /) -> bytes: ...

    async def readany(self) -> bytes: ...

    async def readchunk(self) -> tuple[bytes, bool]: ...

    async def readexactly(self, n: int, /) -> bytes: ...

    def read_nowait(self, n: int = -1, /) -> bytes: ...


@runtime_checkable
class _IOMixinNoOpen(Protocol):
    __slots__ = ()

    @abc.abstractmethod
    def open(self) -> AbstractAsyncContextManager[StreamReaderProto]:
        raise NotImplementedError

    async def read(self, **kwargs: Any) -> bytes:
        """Read the whole contents of this file."""
        async with self.open(**kwargs) as io:
            return await io.read()

    async def save(self, filename: StrOrBytesPath, **kwargs: Any) -> int:
        """Save the file to a path.

        Parameters
        ----------
        filename
            The filename of the file to be created and have this saved to.

        Returns
        -------
        The number of bytes written.
        """
        total = 0
        async with self.open(**kwargs) as file:
            with open(filename, "wb") as actual_fp:  # noqa: PTH123
                async for chunk in file.iter_chunked(2048):
                    total += actual_fp.write(chunk)
        return total

    async def media(self, *, spoiler: bool = False, **kwargs: Any) -> Media:
        """Return this file as :class:`Media` for uploading."""
        return Media(BytesIO(await self.read()), spoiler=spoiler)


@runtime_checkable
class _IOMixin(_IOMixinNoOpen, Protocol):
    __slots__ = ()
    _state: ConnectionState
    url: ReadOnly[str]

    @asynccontextmanager
    async def open(self) -> AsyncGenerator[StreamReaderProto, None]:
        async with self._state.http._session.get(self.url) as r:
            yield r.content


class Avatar(_IOMixin):
    __slots__ = (
        "sha",
        "_suffix",
        "_state",
    )

    def __init__(self, state: ConnectionState, sha: bytes, suffix: str = "full"):
        sha = bytes(sha)
        self.sha = sha if sha != b"\x00" * 20 else DEFAULT_AVATAR
        self._state = state
        self._suffix = suffix

    @property
    def url(self) -> str:
        """The URL of the avatar. Uses the large (184x184 px) image URL."""
        return f"https://avatars.cloudflare.steamstatic.com/{self.sha.hex()}_{self._suffix}.jpg"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__) and self.sha == other.sha

    def __hash__(self) -> int:
        return hash(self.sha)


@dataclass(slots=True, unsafe_hash=True)
class CDNAsset(_IOMixin):
    _state: ConnectionState = field(repr=False, hash=False)
    url: ReadOnly[str] = field()
    """The URL of the asset."""


@dataclass(slots=True)
class Wallet:
    _state: ConnectionState
    balance: int
    """The balance of your wallet in its base most currency denomination.

    E.g. $3.45 -> 345
    """
    currency: Currency
    """The currency the balance is in."""
    balance_delayed: int
    """Your delayed balance if Steam is refunding something?"""
    realm: Realm
    """The realm this wallet is for."""

    async def add(self, code: str) -> int:
        """Add a wallet code to your wallet.

        Parameters
        ----------
        code
            The wallet code to redeem.

        Raises
        ------
        ValueError
            The code was invalid.

        Returns
        -------
        The balance added to your account.
        """
        self._state.handled_wallet.clear()
        resp = await self._state.http.add_wallet_code(code)
        result = Result.try_value(resp["success"])
        if result != Result.OK:
            raise ValueError(
                f"Activation of code failed with result {result} {PurchaseResult.try_value(resp['detail'])!r}"
            )
        await self._state.handled_wallet.wait()
        return resp["amount"]


@runtime_checkable
class AssetMixin(Protocol):
    __slots__ = (
        "_state",
        "_app_id",
        "class_id",
    )

    _state: ConnectionState
    _app_id: AppID
    class_id: ClassID
    """The classid of the item."""

    @property
    def app(self) -> PartialApp[None]:
        """The app the item is from."""
        from .app import PartialApp

        return PartialApp(self._state, id=self._app_id)


@runtime_checkable
class DescriptionMixin(AssetMixin, Protocol):
    __slots__ = SLOTS = (
        "name",
        "type",
        "tags",
        "colour",
        "icon",
        "display_name",
        "market_hash_name",
        "descriptions",
        "owner_descriptions",
        "fraud_warnings",
        "actions",
        "owner_actions",
        "market_actions",
        "market_fee",
        "_market_fee_app_id",
        "_is_tradable",
        "_is_marketable",
    )
    if not TYPE_CHECKING:
        __slots__ = ()

    _app_id: AppID
    class_id: ClassID

    def __init__(self, state: ConnectionState, description: econ.ItemDescription):
        self.name = description.market_name
        """The market_name of the item."""
        self.display_name = description.name or self.name
        """
        The displayed name of the item. This could be different to :attr:`Item.name` if the item is user re-nameable.
        """
        self.market_hash_name = description.market_hash_name or self.name or self.display_name
        """The market_hash_name of the item."""
        self.colour = int(description.name_color, 16) if description.name_color else None
        """The colour of the item."""
        self.descriptions = description.descriptions
        """The descriptions of the item."""
        self.owner_descriptions = description.owner_descriptions
        """The descriptions of the item which are visible only to the owner of the item."""
        self.type = description.type
        """The type of the item."""
        self.tags = description.tags
        """The tags of the item."""
        icon_url = description.icon_url_large or description.icon_url
        self.icon = (
            CDNAsset(state, f"https://community.cloudflare.steamstatic.com/economy/image/{icon_url}")
            if icon_url
            else None
        )
        """The icon url of the item. Uses the large image url where possible."""
        self.fraud_warnings = description.fraudwarnings
        """The fraud warnings for the item."""
        self.actions = description.actions
        """The actions for the item."""
        self.owner_actions = description.owner_actions
        """The owner actions for the item."""
        self.market_actions = description.market_actions
        """The market actions for the item."""
        self.market_fee = (
            int(float(description.market_fee) * 100) if description.market_fee else 10
        )  # if steam ever support currencies that have more than 2 decimals we're screwed
        """The market fee percentage for the item."""
        self._app_id = AppID(description.appid)
        self._market_fee_app_id = AppID(description.market_fee_app)
        self._is_tradable = description.tradable
        self._is_marketable = description.marketable

    def is_tradable(self) -> bool:
        """Whether the item is tradable."""
        return self._is_tradable

    def is_marketable(self) -> bool:
        """Whether the item is marketable."""
        return self._is_marketable

    async def price(self, *, currency: Currency | None = None) -> PriceOverview:
        """Fetch the price of this item on the Steam Community Market place.

        Shorthand for:

        .. code:: python

            await client.fetch_price(item.market_hash_name, item.app, currency)
        """
        price = await self._state.http.get_price(self._app_id, self.market_hash_name, currency)
        return PriceOverview(price, currency or Currency.USD)

    @property
    def market_fee_app(self) -> PartialApp[None]:
        """The app for which the Steam Community Market fee percentage is applied."""
        from .app import PartialApp

        return PartialApp(self._state, id=self._market_fee_app_id)



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any, Literal, TypeVar

from multidict import MultiDict

if TYPE_CHECKING:
    from typing import TypeAlias

    from typing_extensions import Never, TypedDict

# VDF related types
VDFDict: TypeAlias = MultiDict["str | VDFDict"]  # str | bool | None | VDFDict for orvdf
BinaryVDFDict: TypeAlias = MultiDict[
    "str | int | float | BinaryVDFDict"
]  # str | int | float | bool | None | BinaryVDFDict for orvdf


T = TypeVar("T")
VDFInt: TypeAlias = str
VDFBool: TypeAlias = Literal["0", "1"]


class VDFList(MultiDict[T]):
    # lists don't strictly exist in the VDF "spec" however Valve uses {"0": T, "1": T, ..., "n": T}
    # to represent them, so prevent people doing daft things like use .keys() and iterating over it
    # for now using __getitem__ is fine, but I might remove this at some point

    def keys(self) -> Never: ...

    def __iter__(self) -> Never: ...


if TYPE_CHECKING:

    class TypedVDFDict(MultiDict[Any], TypedDict):  # type: ignore  # this obviously doesn't work at runtime
        pass

else:

    class TypedVDFDict(MultiDict[Any]):
        def __init_subclass__(cls, **kwargs: Any) -> None:
            pass



"""
steam.ext.csgo
~~~~~~~~~~~~~~

A library for interacting with the CSGO Game Coordinator.

Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE
"""

from . import utils as utils
from .backpack import *
from .client import *
from .enums import *
from .models import *



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING, Any

from .abc import Channel
from .chat import Chat, GroupChannelProtos
from .message import ClanMessage, GroupMessage

if TYPE_CHECKING:
    from collections.abc import AsyncGenerator, Coroutine
    from contextlib import AbstractAsyncContextManager
    from datetime import datetime

    from .clan import Clan
    from .friend import Friend
    from .group import Group
    from .media import Media
    from .message import UserMessage
    from .state import ConnectionState
    from .user import User


__all__ = (
    "UserChannel",
    "GroupChannel",
    "ClanChannel",
)


class UserChannel(Channel["UserMessage", None, None]):
    """Represents the channel a DM is sent in."""

    __slots__ = ("participant", "last_message")
    clan: None
    group: None

    def __init__(self, state: ConnectionState, participant: User | Friend):
        super().__init__(state)
        self.participant = participant
        """The recipient of any messages sent."""
        self.last_message: UserMessage | None = None

    def __repr__(self) -> str:
        return f"<{self.__class__.__name__} participant={self.participant!r}>"

    def __eq__(self, other: object) -> bool:
        return isinstance(other, UserChannel) and self.participant == other.participant

    def __hash__(self) -> int:
        return hash(self.participant)

    def _send_message(self, content: str) -> Coroutine[Any, Any, UserMessage]:
        return self.participant._send_message(content)

    def _send_media(self, media: Media) -> Coroutine[Any, Any, None]:
        return self.participant._send_media(media)

    def typing(self) -> AbstractAsyncContextManager[None]:
        """Send a typing indicator continuously to the channel while in the context manager.

        Note
        ----
        This only works in DMs.

        Usage:

        .. code:: python

            async with channel.typing():
                ...  # do your expensive operations
        """

        return self.participant.typing()

    async def trigger_typing(self) -> None:
        """Send a typing indicator to the channel once.

        Note
        ----
        This only works in DMs.
        """
        await self.participant.trigger_typing()

    def history(
        self,
        *,
        limit: int | None = 100,
        before: datetime | None = None,
        after: datetime | None = None,
    ) -> AsyncGenerator[UserMessage, None]:
        return self.participant.history(limit=limit, before=before, after=after)


class GroupChannel(Chat[GroupMessage, None, "Group"]):
    """Represents a group channel."""

    __slots__ = ()

    def __init__(self, state: ConnectionState, group: Group, proto: GroupChannelProtos):
        super().__init__(state, group, proto)  # type: ignore  # cbf fixing this
        self.group = group


class ClanChannel(Chat[ClanMessage, "Clan", None]):
    """Represents a clan channel."""

    __slots__ = ()

    def __init__(self, state: ConnectionState, clan: Clan, proto: GroupChannelProtos):
        super().__init__(state, clan, proto)  # type: ignore
        self.clan = clan



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

import inspect
import sys
import traceback
from collections.abc import Callable, Mapping
from copy import deepcopy
from typing import TYPE_CHECKING, Any, Concatenate, Final, Generic, TypeAlias, overload

from typing_extensions import TypeVar

from .commands import CogT, Command, Group, check
from .utils import Coro

if TYPE_CHECKING:
    from steam.ext import commands

    from .bot import Bot

__all__ = ("Cog",)

ListenerType: TypeAlias = Callable[..., Coro[None]]
UnboundListenerType: TypeAlias = "Callable[Concatenate[CogT, ...], Coro[None]]"
F = TypeVar("F", bound=UnboundListenerType)


class CogCommands:
    @overload
    def __get__(self, instance: None, owner: type[Cog]) -> set[Command[None]]: ...

    @overload
    def __get__(self, instance: CogT, owner: type[CogT]) -> set[Command[CogT]]:  # type: ignore
        ...

    def __get__(self, instance: Cog | None, owner: type[Cog]) -> Any:
        commands = owner.__commands__
        if instance is None:
            return {command for command in commands.values() if command.parent is None}
        cog_qualified_names = {command.qualified_name for command in commands.values()}
        return {command for command in instance.bot.commands if command.qualified_name in cog_qualified_names}


class CogListeners:
    @overload
    def __get__(self, instance: None, owner: type[CogT]) -> list[tuple[str, UnboundListenerType[CogT]]]:  # type: ignore
        ...

    @overload
    def __get__(self, instance: Cog, owner: type[Cog]) -> list[tuple[str, ListenerType]]: ...

    def __get__(self, instance: Cog | None, owner: type[Cog]) -> Any:
        if instance is None:
            return [
                (listener.__name__, listener) for listeners in owner.__listeners__.values() for listener in listeners
            ]

        return [
            (listener.__name__, getattr(instance, listener.__name__))
            for listeners in owner.__listeners__.values()
            for listener in listeners
        ]


BotT = TypeVar("BotT", bound="Bot", default="Bot", covariant=True)


class Cog(Generic[BotT]):
    """A class from which Cogs can be created. These are used to separate commands and listeners into separate files.

    Attributes
    ----------
    qualified_name
        The name of the cog. Can be set in a subclass e.g.

        .. code:: python

            class MyCog(commands.Cog, name="SpecialCogName"): ...

        Defaults to ``Cog.__name__``.

    command_attrs
        Attributes to pass to every command registered in the cog.
        Can be set in subclass e.g.

        .. code:: python

            class MyBrokenCog(commands.Cog, command_attrs=dict(enabled=False)):
                # all the commands by default would be disabled

                @commands.command()
                async def broken(self, ctx):  # disabled
                    ...

                @commands.command(enabled=True)
                async def working(self, ctx):  # enabled
                    ...

    description
        The cleaned up doc-string for the cog.
    """

    __commands__: Final[Mapping[str, Command[None]]]
    __listeners__: Final[Mapping[str, list[UnboundListenerType]]]
    command_attrs: Final[Mapping[str, Any]]
    qualified_name: Final[str]
    description: Final[str | None]
    bot: Final[BotT]

    def __init_subclass__(
        cls, name: str | None = None, command_attrs: dict[str, Any] | None = None, **kwargs: Any
    ) -> None:
        cls.qualified_name = name or cls.__name__  # type: ignore
        cls.command_attrs = command_attrs or {}  # type: ignore
        cls.description = inspect.cleandoc(cls.__doc__) if cls.__doc__ is not None else None  # type: ignore

        cls.__listeners__ = {}  # type: ignore
        cls.__commands__ = {}  # type: ignore
        for name, attr in inspect.getmembers(cls):
            if name.startswith(("bot_", "cog_")) and getattr(Cog, name, None) is None:
                raise TypeError(
                    (
                        f'Methods prefixed with "bot_" or "cog_" are reserved for future use, {attr.__qualname__} is '
                        f"therefore not allowed"
                    )
                )
            if isinstance(attr, Command):
                if isinstance(attr, Group):
                    for name, value in cls.command_attrs.items():
                        for child in attr.children:
                            if name not in child.__original_kwargs__:
                                setattr(child, name, value)
                else:
                    cls.__commands__[name] = attr  # type: ignore

                for name, value in cls.command_attrs.items():
                    if name not in attr.__original_kwargs__:
                        setattr(attr, name, value)

            elif hasattr(attr, "__event_name__"):
                try:
                    cls.__listeners__[attr.__event_name__].append(attr)
                except KeyError:
                    cls.__listeners__[attr.__event_name__] = [attr]
        super().__init_subclass__(**kwargs)

    def __repr__(self) -> str:
        return f"<Cog {f'{self.__class__.__module__}.{self.__class__.__name__}'!r}>"

    commands: Final = CogCommands()
    """A set of the cog's commands."""

    listeners: Final = CogListeners()
    """A list of tuples of the events registered with the format (name, listener)"""

    @classmethod
    @overload
    def listener(cls, coro: F, /) -> F: ...

    @classmethod
    @overload
    def listener(cls, name: str | None = None, /) -> Callable[[F], F]: ...

    @classmethod
    def listener(cls, name: F | str | None = None, /) -> Callable[[F], F] | F:
        """|maybecallabledeco|
        Register a :term:`coroutine function` as a listener. Similar to :meth:`~steam.ext.commands.Bot.listen`.

        Parameters
        ----------
        name
            The name of the event to listen for. Defaults to ``func.__name__``.
        """

        def decorator(coro: F) -> F:
            if not inspect.iscoroutinefunction(coro):
                raise TypeError(f"Listeners must be coroutine functions, {coro.__name__} is {type(coro).__name__}")
            coro.__event_name__ = coro.__name__ if callable(name) else name  # type: ignore
            return coro

        return decorator(name) if callable(name) else decorator

    async def cog_command_error(self, ctx: commands.Context[Bot], error: Exception) -> None:
        """A special method that is called when an error is dispatched inside this cog. This is similar to
        :func:`~commands.Bot.on_command_error` except it only applies to the commands inside this cog.

        Parameters
        -----------
        ctx
            The invocation context where the error happened.
        error
            The error that happened.
        """
        print(f"Ignoring exception in command {ctx.command}:", file=sys.stderr)
        traceback.print_exception(type(error), error, error.__traceback__, file=sys.stderr)

    async def cog_check(self, ctx: commands.Context[Bot]) -> bool:
        """A special method that registers as a :func:`commands.check` for every command and subcommand in this cog.

        Parameters
        -----------
        ctx
            The invocation context.
        """
        raise NotImplementedError

    async def bot_check(self, ctx: commands.Context[Bot]) -> bool:
        """A special method that registers as a :meth:`commands.Bot.check` globally.

        Parameters
        -----------
        ctx
            The invocation context.
        """
        raise NotImplementedError

    async def cog_unload(self) -> None:
        """A special method that is called when the cog gets removed.

        This is called before :func:`teardown`.
        """

    async def _inject(self, bot: BotT) -> None:  # type: ignore
        cls = self.__class__
        self.bot = bot  # type: ignore
        if cls.bot_check is not Cog.bot_check:
            bot.add_check(check(self.bot_check))

        for idx, command in enumerate(self.__commands__.values()):
            command = deepcopy(command)
            command.cog = self

            if cls.cog_check is not Cog.cog_check:
                command.checks.append(check(self.cog_check))

            if isinstance(command, Group):
                for child in command.children:
                    child.cog = self

            try:
                bot.add_command(command)
            except Exception:
                # undo our additions
                for to_undo in tuple(self.__commands__)[:idx]:
                    bot.remove_command(to_undo)
                raise

        for name, listener in self.listeners:
            bot.add_listener(listener, name)

    async def _eject(self, bot: Bot) -> None:
        for command in self.__commands__.values():
            if isinstance(command, Group):
                command.remove_all_commands()
            else:
                bot.remove_command(command.name)

        for name, listener in self.listeners:
            bot.remove_listener(listener, name)

        await self.cog_unload()



"""Licensed under The MIT License (MIT) - Copyright (c) 2020-present James H-B. See LICENSE"""

from __future__ import annotations

from typing import TYPE_CHECKING

from ...enums import Flags, IntEnum, classproperty

if TYPE_CHECKING:
    from collections.abc import Mapping

    from typing_extensions import Self

__all__ = (
    "ItemQuality",
    "ItemFlags",
    "ItemOrigin",
    "Rank",
)


# fmt: off
class ItemCustomizationNotification(IntEnum):
    NameItem              = 1006
    UnlockCrate           = 1007
    XRayItemReveal        = 1008
    XRayItemClaim         = 1009
    CasketTooFull         = 1011
    CasketContents        = 1012
    CasketAdded           = 1013
    CasketRemoved         = 1014
    CasketInvFull         = 1015
    NameBaseItem          = 1019
    RemoveItemName        = 1030
    RemoveSticker         = 1053
    ApplySticker          = 1086
    StatTrakSwap          = 1088
    ActivateFanToken      = 9178
    ActivateOperationCoin = 9179
    GraffitiUnseal        = 9185
    GenerateSouvenir      = 9204


# All 3 lifted from
# https://github.com/perilouswithadollarsign/cstrike15_src/blob/f82112a2388b841d72cb62ca48ab1846dfcc11c8/game/shared/econ/econ_item_constants.h
class ItemQuality(IntEnum):
    Undefined  = -1
    Normal     = 0
    Genuine    = 1
    Vintage    = 2
    Unusual    = 3
    Unique     = 4
    Community  = 5
    Developer  = 6
    SelfMade   = 7
    Customised = 8
    Strange    = 9
    Completed  = 10
    Haunted    = 11
    Tournament = 12
    Favoured   = 13
    Max        = 14


class ItemFlags(Flags):
    CannotTrade                                = 1 << 0
    CannotCraft                                = 1 << 1
    CanBeTradedByFreeAccounts                  = 1 << 2
    NotEcon                                    = 1 << 3
    """Items that cannot interact in the economy (can't be traded, gift-wrapped, crafted, etc.)"""
    PurchasedAfterStoreCraftabilityChanges2012 = 1 << 4
    """Cosmetic items coming from the store are now usable in crafting"""
    ForceBlueTeam                              = 1 << 5
    StoreItem                                  = 1 << 6
    Preview                                    = 1 << 7


class ItemOrigin(IntEnum):
    Invalid                   = -1
    Drop                      = 0
    Achievement               = 1
    Purchased                 = 2
    Traded                    = 3
    Crafted                   = 4
    StorePromotion            = 5
    Gifted                    = 6
    SupportGranted            = 7
    FoundInCrate              = 8
    Earned                    = 9
    ThirdPartyPromotion       = 10
    GiftWrapped               = 11
    HalloweenDrop             = 12
    PackageItem               = 13
    Foreign                   = 14
    CDKey                     = 15
    CollectionReward          = 16
    PreviewItem               = 17
    SteamWorkshopContribution = 18
    PeriodicScoreReward       = 19
    Recycling                 = 20
    TournamentDrop            = 21
    StockItem                 = 22
    QuestReward               = 23
    LevelUpReward             = 24
    Max                       = 25



class Rank(IntEnum):
    NotRanked                   = 0
    SilverI                     = 1
    SilverII                    = 2
    SilverIII                   = 3
    SilverIV                    = 4
    SilverElite                 = 5
    SilverEliteMaster           = 6
    GoldNovaI                   = 7
    GoldNovaII                  = 8
    GoldNovaIII                 = 9
    GoldNovaMaster              = 10
    MasterGuardianI             = 11
    MasterGuardianII            = 12
    MasterGuardianElite         = 13
    DistinguishedMasterGuardian = 14
    LegendaryEagle              = 15
    LegendaryEagleMaster        = 16
    SupremeMasterFirstClass     = 17
    TheGlobalElite              = 18

    @classproperty
    def DISPLAY_NAMES(cls: type[Self]) -> Mapping[Rank, str]:  # type: ignore
        return {
            cls.NotRanked:                   "Not Ranked",
            cls.SilverI:                     "Silver I",
            cls.SilverII:                    "Silver II",
            cls.SilverIII:                   "Silver III",
            cls.SilverIV:                    "Silver IV",
            cls.SilverElite:                 "Silver Elite",
            cls.SilverEliteMaster:           "Silver Elite Master",
            cls.GoldNovaI:                   "Gold Nova I",
            cls.GoldNovaII:                  "Gold Nova II",
            cls.GoldNovaIII:                 "Gold Nova III",
            cls.GoldNovaMaster:              "Gold Nova Master",
            cls.MasterGuardianI:             "Master Guardian I",
            cls.MasterGuardianII:            "Master Guardian II",
            cls.MasterGuardianElite:         "Master Guardian Elite",
            cls.DistinguishedMasterGuardian: "Distinguished Master Guardian",
            cls.LegendaryEagle:              "Legendary Eagle",
            cls.LegendaryEagleMaster:        "Legendary Eagle Master",
            cls.SupremeMasterFirstClass:     "Supreme Master First Class",
            cls.TheGlobalElite:              "The Global Elite",
        }

    @property
    def display_name(self) -> str:
        return self.DISPLAY_NAMES[self]

class RankType(IntEnum):
    Matchmaking          = 6
    Wingman              = 7
    CS2                  = 10
    Dangerzone           = 11

class EMsg(IntEnum):
    # ESOMsg
    SOCreate                               = 21
    SOUpdate                               = 22
    SODestroy                              = 23
    SOCacheSubscribed                      = 24
    SOCacheUnsubscribed                    = 25
    SOUpdateMultiple                       = 26
    SOCacheSubscriptionCheck               = 27
    SOCacheSubscriptionRefresh             = 28

    # EGCItemMsg
    EGCItemMsgBase                         = 1000
    SetItemPosition                        = 1001
    Craft                                  = 1002
    CraftResponse                          = 1003
    Delete                                 = 1004
    VerifyCacheSubscription                = 1005
    NameItem                               = 1006
    UnlockCrate                            = 1007
    UnlockCrateResponse                    = 1008
    PaintItem                              = 1009
    PaintItemResponse                      = 1010
    GoldenWrenchBroadcast                  = 1011
    MOTDRequest                            = 1012
    MOTDRequestResponse                    = 1013
    NameBaseItem                           = 1019
    NameBaseItemResponse                   = 1020
    CustomizeItemTexture                   = 1023
    CustomizeItemTextureResponse           = 1024
    UseItemRequest                         = 1025
    UseItemResponse                        = 1026
    RemoveItemName                         = 1030
    RemoveItemPaint                        = 1031
    GiftWrapItem                           = 1032
    GiftWrapItemResponse                   = 1033
    DeliverGift                            = 1034
    DeliverGiftResponseGiver               = 1035
    DeliverGiftResponseReceiver            = 1036
    UnwrapGiftRequest                      = 1037
    UnwrapGiftResponse                     = 1038
    SetItemStyle                           = 1039
    UsedClaimCodeItem                      = 1040
    SortItems                              = 1041
    LookupAccount                          = 1043
    LookupAccountResponse                  = 1044
    LookupAccountName                      = 1045
    LookupAccountNameResponse              = 1046
    UpdateItemSchema                       = 1049
    RemoveCustomTexture                    = 1051
    RemoveCustomTextureResponse            = 1052
    RemoveMakersMark                       = 1053
    RemoveMakersMarkResponse               = 1054
    RemoveUniqueCraftIndex                 = 1055
    RemoveUniqueCraftIndexResponse         = 1056
    SaxxyBroadcast                         = 1057
    BackpackSortFinished                   = 1058
    AdjustItemEquippedState                = 1059
    CollectItem                            = 1061
    PresetsSelectPresetForClass            = 1063
    PresetsSetItemPosition                 = 1064
    ReportAbuse                            = 1065
    ReportAbuseResponse                    = 1066
    PresetsSelectPresetForClassReply       = 1067
    NameItemNotification                   = 1068
    ApplyConsumableEffects                 = 1069
    ConsumableExhausted                    = 1070
    ShowItemsPickedUp                      = 1071
    ClientDisplayNotification              = 1072
    ApplyStrangePart                       = 1073
    IncrementKillCountAttribute            = 1074
    IncrementKillCountResponse             = 1075
    ApplyPennantUpgrade                    = 1076
    SetItemPositions                       = 1077
    ApplyEggEssence                        = 1078
    NameEggEssenceResponse                 = 1079
    PaintKitItem                           = 1080
    PaintKitBaseItem                       = 1081
    PaintKitItemResponse                   = 1082
    GiftedItems                            = 1083
    UnlockItemStyle                        = 1084
    UnlockItemStyleResponse                = 1085
    ApplySticker                           = 1086
    ItemAcknowledged                       = 1087
    StatTrakSwap                           = 1088
    UserTrackTimePlayedConsecutively       = 1089
    ItemCustomizationNotification          = 1090
    ModifyItemAttribute                    = 1091
    CasketItemAdd                          = 1092
    CasketItemExtract                      = 1093
    CasketItemLoadContents                 = 1094
    TradingBase                            = 1500
    TradingInitiateTradeRequest            = 1501
    TradingInitiateTradeResponse           = 1502
    TradingStartSession                    = 1503
    TradingSetItem                         = 1504
    TradingRemoveItem                      = 1505
    TradingUpdateTradeInfo                 = 1506
    TradingSetReadiness                    = 1507
    TradingReadinessResponse               = 1508
    TradingSessionClosed                   = 1509
    TradingCancelSession                   = 1510
    TradingTradeChatMsg                    = 1511
    TradingConfirmOffer                    = 1512
    TradingTradeTypingChatMsg              = 1513
    ServerBrowserFavoriteServer            = 1601
    ServerBrowserBlacklistServer           = 1602
    ServerRentalsBase                      = 1700
    ItemPreviewCheckStatus                 = 1701
    ItemPreviewStatusResponse              = 1702
    ItemPreviewRequest                     = 1703
    ItemPreviewRequestResponse             = 1704
    ItemPreviewExpire                      = 1705
    ItemPreviewExpireNotification          = 1706
    ItemPreviewItemBoughtNotification      = 1707
    DevNewItemRequest                      = 2001
    DevNewItemRequestResponse              = 2002
    DevPaintKitDropItem                    = 2003
    StoreGetUserData                       = 2500
    StoreGetUserDataResponse               = 2501
    StorePurchaseFinalize                  = 2504
    StorePurchaseFinalizeResponse          = 2505
    StorePurchaseCancel                    = 2506
    StorePurchaseCancelResponse            = 2507
    StorePurchaseQueryTxn                  = 2508
    StorePurchaseQueryTxnResponse          = 2509
    StorePurchaseInit                      = 2510
    StorePurchaseInitResponse              = 2511
    BannedWordListRequest                  = 2512
    BannedWordListResponse                 = 2513
    GCToGCBannedWordListBroadcast          = 2514
    GCToGCBannedWordListUpdated            = 2515
    GCToGCDirtySDOCache                    = 2516
    GCToGCDirtyMultipleSDOCache            = 2517
    GCToGCUpdateSQLKeyValue                = 2518
    GCToGCIsTrustedServer                  = 2519
    GCToGCIsTrustedServerResponse          = 2520
    GCToGCBroadcastConsoleCommand          = 2521
    ServerVersionUpdated                   = 2522
    ApplyAutograph                         = 2523
    GCToGCWebAPIAccountChanged             = 2524
    RequestAnnouncements                   = 2525
    RequestAnnouncementsResponse           = 2526
    RequestPassportItemGrant               = 2527
    ClientVersionUpdated                   = 2528

    # EGCBaseClientMsg
    ClientWelcome                          = 4004
    ServerWelcome                          = 4005
    ClientHello                            = 4006
    ServerHello                            = 4007
    ClientConnectionStatus                 = 4009
    ServerConnectionStatus                 = 4010

    # ECsgoGCMsg
    ECsgoGCMsgBase                         = 9100
    MatchmakingStart                       = 9101
    MatchmakingStop                        = 9102
    MatchmakingClient2ServerPing           = 9103
    MatchmakingGC2ClientUpdate             = 9104
    MatchmakingGC2ServerReserve            = 9105
    MatchmakingServerReservationResponse   = 9106
    MatchmakingGC2ClientReserve            = 9107
    MatchmakingServerRoundStats            = 9108
    MatchmakingClient2GCHello              = 9109
    MatchmakingGC2ClientHello              = 9110
    MatchmakingServerMatchEnd              = 9111
    MatchmakingGC2ClientAbandon            = 9112
    MatchmakingServer2GCKick               = 9113
    MatchmakingGC2ServerConfirm            = 9114
    MatchmakingGCOperationalStats          = 9115
    MatchmakingGC2ServerRankUpdate         = 9116
    MatchmakingOperator2GCBlogUpdate       = 9117
    ServerNotificationForUserPenalty       = 9118
    ClientReportPlayer                     = 9119
    ClientReportServer                     = 9120
    ClientCommendPlayer                    = 9121
    ClientReportResponse                   = 9122
    ClientCommendPlayerQuery               = 9123
    ClientCommendPlayerQueryResponse       = 9124
    WatchInfoUsers                         = 9126
    ClientRequestPlayersProfile            = 9127
    PlayersProfile                         = 9128
    SetMyMedalsInfo                        = 9129
    PlayerEarnedRewardNotification         = 9130
    PlayerOverwatchCaseUpdate              = 9131
    PlayerOverwatchCaseAssignment          = 9132
    PlayerOverwatchCaseStatus              = 9133
    GC2ClientTextMsg                       = 9134
    Client2GCTextMsg                       = 9135
    MatchEndRunRewardDrops                 = 9136
    MatchEndRewardDropsNotification        = 9137
    ClientRequestWatchInfoFriends2         = 9138
    MatchList                              = 9139
    MatchListRequestCurrentLiveGames       = 9140
    MatchListRequestRecentUserGames        = 9141
    GC2ServerReservationUpdate             = 9142
    ClientVarValueNotificationInfo         = 9144
    TournamentMatchRewardDropsNotification = 9145
    MatchListRequestTournamentGames        = 9146
    MatchListRequestFullGameInfo           = 9147
    GiftsLeaderboardRequest                = 9148
    GiftsLeaderboardResponse               = 9149
    ServerVarValueNotificationInfo         = 9150
    ClientSubmitSurveyVote                 = 9152
    Server2GCClientValidate                = 9153
    MatchListRequestLiveGameForUser        = 9154
    Server2GCPureServerValidationFailure   = 9155
    Client2GCEconPreviewDataBlockRequest   = 9156
    Client2GCEconPreviewDataBlockResponse  = 9157
    AccountPrivacySettings                 = 9158
    SetMyActivityInfo                      = 9159
    MatchListRequestTournamentPredictions  = 9160
    MatchListUploadTournamentPredictions   = 9161
    DraftSummary                           = 9162
    ClientRequestJoinFriendData            = 9163
    ClientRequestJoinServerData            = 9164
    ClientRequestNewMission                = 9165
    PremierSeasonSummary                   = 9224



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: base_gcmessages.proto
# plugin: python-betterproto

from dataclasses import dataclass

import betterproto

from ....protobufs.msg import GCProtobufMessage
from ..enums import EMsg


class EgcBaseProtoObjectTypes(betterproto.Enum):
    PartyInvite = 1001
    LobbyInvite = 1002


class GcBannedWordType(betterproto.Enum):
    DisableWord = 0
    EnableWord = 1


class GameServerInfoServerType(betterproto.Enum):
    Unspecified = 0
    Game = 1
    Proxy = 2


@dataclass(eq=False, repr=False)
class CgcStorePurchaseInitLineItem(betterproto.Message):
    item_def_id: int = betterproto.uint32_field(1)
    quantity: int = betterproto.uint32_field(2)
    cost_in_local_currency: int = betterproto.uint32_field(3)
    purchase_type: int = betterproto.uint32_field(4)


class StorePurchaseInit(GCProtobufMessage, msg=EMsg.StorePurchaseInit):
    country: str = betterproto.string_field(1)
    language: int = betterproto.int32_field(2)
    currency: int = betterproto.int32_field(3)
    line_items: "list[CgcStorePurchaseInitLineItem]" = betterproto.message_field(4)


@dataclass(eq=False, repr=False)
class StorePurchaseInitResponse(betterproto.Message):
    result: int = betterproto.int32_field(1)
    txn_id: int = betterproto.uint64_field(2)
    url: str = betterproto.string_field(3)
    item_ids: list[int] = betterproto.uint64_field(4)


@dataclass(eq=False, repr=False)
class CsoPartyInvite(betterproto.Message):
    group_id: int = betterproto.uint64_field(1)
    sender_id: int = betterproto.fixed64_field(2)
    sender_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class CsoLobbyInvite(betterproto.Message):
    group_id: int = betterproto.uint64_field(1)
    sender_id: int = betterproto.fixed64_field(2)
    sender_name: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class SystemBroadcast(betterproto.Message):
    message: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class InviteToParty(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    client_version: int = betterproto.uint32_field(2)
    team_invite: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class InvitationCreated(betterproto.Message):
    group_id: int = betterproto.uint64_field(1)
    steam_id: int = betterproto.fixed64_field(2)


@dataclass(eq=False, repr=False)
class PartyInviteResponse(betterproto.Message):
    party_id: int = betterproto.uint64_field(1)
    accept: bool = betterproto.bool_field(2)
    client_version: int = betterproto.uint32_field(3)
    team_invite: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class KickFromParty(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class LeaveParty(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ServerAvailable(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class LanServerAvailable(betterproto.Message):
    lobby_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class GameAccountClient(betterproto.Message):
    additional_backpack_slots: int = betterproto.uint32_field(1)
    bonus_xp_timestamp_refresh: int = betterproto.fixed32_field(12)
    bonus_xp_usedflags: int = betterproto.uint32_field(13)
    elevated_state: int = betterproto.uint32_field(14)
    elevated_timestamp: int = betterproto.uint32_field(15)


@dataclass(eq=False, repr=False)
class CriteriaCondition(betterproto.Message):
    op: int = betterproto.int32_field(1)
    field: str = betterproto.string_field(2)
    required: bool = betterproto.bool_field(3)
    float_value: float = betterproto.float_field(4)
    string_value: str = betterproto.string_field(5)


@dataclass(eq=False, repr=False)
class Criteria(betterproto.Message):
    item_level: int = betterproto.uint32_field(1)
    item_quality: int = betterproto.int32_field(2)
    item_level_set: bool = betterproto.bool_field(3)
    item_quality_set: bool = betterproto.bool_field(4)
    initial_inventory: int = betterproto.uint32_field(5)
    initial_quantity: int = betterproto.uint32_field(6)
    ignore_enabled_flag: bool = betterproto.bool_field(8)
    conditions: "list[CriteriaCondition]" = betterproto.message_field(9)
    item_rarity: int = betterproto.int32_field(10)
    item_rarity_set: bool = betterproto.bool_field(11)
    recent_only: bool = betterproto.bool_field(12)


@dataclass(eq=False, repr=False)
class Recipe(betterproto.Message):
    def_index: int = betterproto.uint32_field(1)
    name: str = betterproto.string_field(2)
    n_a: str = betterproto.string_field(3)
    desc_inputs: str = betterproto.string_field(4)
    desc_outputs: str = betterproto.string_field(5)
    di_a: str = betterproto.string_field(6)
    di_b: str = betterproto.string_field(7)
    di_c: str = betterproto.string_field(8)
    do_a: str = betterproto.string_field(9)
    do_b: str = betterproto.string_field(10)
    do_c: str = betterproto.string_field(11)
    requires_all_same_class: bool = betterproto.bool_field(12)
    requires_all_same_slot: bool = betterproto.bool_field(13)
    class_usage_for_output: int = betterproto.int32_field(14)
    slot_usage_for_output: int = betterproto.int32_field(15)
    set_for_output: int = betterproto.int32_field(16)
    input_items_criteria: "list[Criteria]" = betterproto.message_field(20)
    output_items_criteria: "list[Criteria]" = betterproto.message_field(21)
    input_item_dupe_counts: list[int] = betterproto.uint32_field(22)


@dataclass(eq=False, repr=False)
class DevNewItemRequest(betterproto.Message):
    receiver: int = betterproto.fixed64_field(1)
    criteria: "Criteria" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class IncrementKillCountAttribute(betterproto.Message):
    killer_account_id: int = betterproto.fixed32_field(1)
    victim_account_id: int = betterproto.fixed32_field(2)
    item_id: int = betterproto.uint64_field(3)
    event_type: int = betterproto.uint32_field(4)
    amount: int = betterproto.uint32_field(5)


@dataclass(eq=False, repr=False)
class ApplySticker(betterproto.Message):
    sticker_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)
    sticker_slot: int = betterproto.uint32_field(3)
    baseitem_defidx: int = betterproto.uint32_field(4)
    sticker_wear: float = betterproto.float_field(5)


@dataclass(eq=False, repr=False)
class ModifyItemAttribute(betterproto.Message):
    item_id: int = betterproto.uint64_field(1)
    attr_defidx: int = betterproto.uint32_field(2)
    attr_value: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class ApplyStatTrakSwap(betterproto.Message):
    tool_item_id: int = betterproto.uint64_field(1)
    item_1_item_id: int = betterproto.uint64_field(2)
    item_2_item_id: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class ApplyStrangePart(betterproto.Message):
    strange_part_item_id: int = betterproto.uint64_field(1)
    item_item_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ApplyPennantUpgrade(betterproto.Message):
    upgrade_item_id: int = betterproto.uint64_field(1)
    pennant_item_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ApplyEggEssence(betterproto.Message):
    essence_item_id: int = betterproto.uint64_field(1)
    egg_item_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ItemAttribute(betterproto.Message):
    def_index: int = betterproto.uint32_field(1)
    value: int = betterproto.uint32_field(2)
    value_bytes: bytes = betterproto.bytes_field(3)


@dataclass(eq=False, repr=False)
class ItemEquipped(betterproto.Message):
    new_class: int = betterproto.uint32_field(1)
    new_slot: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class Item(betterproto.Message):
    id: int = betterproto.uint64_field(1)
    account_id: int = betterproto.uint32_field(2)
    inventory: int = betterproto.uint32_field(3)
    def_index: int = betterproto.uint32_field(4)
    quantity: int = betterproto.uint32_field(5)
    level: int = betterproto.uint32_field(6)
    quality: int = betterproto.uint32_field(7)
    flags: int = betterproto.uint32_field(8)
    origin: int = betterproto.uint32_field(9)
    custom_name: str = betterproto.string_field(10)
    custom_description: str = betterproto.string_field(11)
    attribute: "list[ItemAttribute]" = betterproto.message_field(12)
    interior_item: "Item" = betterproto.message_field(13)
    in_use: bool = betterproto.bool_field(14)
    style: int = betterproto.uint32_field(15)
    original_id: int = betterproto.uint64_field(16)
    equipped_state: "list[ItemEquipped]" = betterproto.message_field(18)
    rarity: int = betterproto.uint32_field(19)


@dataclass(eq=False, repr=False)
class AdjustItemEquippedState(betterproto.Message):
    item_id: int = betterproto.uint64_field(1)
    new_class: int = betterproto.uint32_field(2)
    new_slot: int = betterproto.uint32_field(3)
    swap: bool = betterproto.bool_field(4)


@dataclass(eq=False, repr=False)
class AdjustItemEquippedStateMulti(betterproto.Message):
    t_equips: list[int] = betterproto.uint64_field(1)
    ct_equips: list[int] = betterproto.uint64_field(2)
    noteam_equips: list[int] = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class SortItems(betterproto.Message):
    sort_type: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ClaimCode(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    code_type: int = betterproto.uint32_field(2)
    time_acquired: int = betterproto.uint32_field(3)
    code: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class StoreGetUserData(betterproto.Message):
    price_sheet_version: int = betterproto.fixed32_field(1)
    currency: int = betterproto.int32_field(2)


@dataclass(eq=False, repr=False)
class StoreGetUserDataResponse(betterproto.Message):
    result: int = betterproto.int32_field(1)
    currency_deprecated: int = betterproto.int32_field(2)
    country_deprecated: str = betterproto.string_field(3)
    price_sheet_version: int = betterproto.fixed32_field(4)
    price_sheet: bytes = betterproto.bytes_field(8)


@dataclass(eq=False, repr=False)
class UpdateItemSchema(betterproto.Message):
    items_game: bytes = betterproto.bytes_field(1)
    item_schema_version: int = betterproto.fixed32_field(2)
    items_game_url_deprecated2013: str = betterproto.string_field(3)
    items_game_url: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class Error(betterproto.Message):
    error_text: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class RequestInventoryRefresh(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ConVarValue(betterproto.Message):
    name: str = betterproto.string_field(1)
    value: str = betterproto.string_field(2)


@dataclass(eq=False, repr=False)
class ReplicateConVars(betterproto.Message):
    convars: "list[ConVarValue]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class UseItem(betterproto.Message):
    item_id: int = betterproto.uint64_field(1)
    target_steam_id: int = betterproto.fixed64_field(2)
    gift_potential_targets: list[int] = betterproto.uint32_field(3)
    duel_class_lock: int = betterproto.uint32_field(4)
    initiator_steam_id: int = betterproto.fixed64_field(5)


@dataclass(eq=False, repr=False)
class ReplayUploadedToYouTube(betterproto.Message):
    youtube_url: str = betterproto.string_field(1)
    youtube_account_name: str = betterproto.string_field(2)
    session_id: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class ConsumableExhausted(betterproto.Message):
    item_def_id: int = betterproto.int32_field(1)


@dataclass(eq=False, repr=False)
class ItemAcknowledgedDeprecated(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    inventory: int = betterproto.uint32_field(2)
    def_index: int = betterproto.uint32_field(3)
    quality: int = betterproto.uint32_field(4)
    rarity: int = betterproto.uint32_field(5)
    origin: int = betterproto.uint32_field(6)
    item_id: int = betterproto.uint64_field(7)


@dataclass(eq=False, repr=False)
class SetItemPositions(betterproto.Message):
    item_positions: "list[SetItemPositionsItemPosition]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class SetItemPositionsItemPosition(betterproto.Message):
    legacy_item_id: int = betterproto.uint32_field(1)
    position: int = betterproto.uint32_field(2)
    item_id: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class ReportAbuse(betterproto.Message):
    target_steam_id: int = betterproto.fixed64_field(1)
    description: str = betterproto.string_field(4)
    gid: int = betterproto.uint64_field(5)
    abuse_type: int = betterproto.uint32_field(2)
    content_type: int = betterproto.uint32_field(3)
    target_game_server_ip: int = betterproto.fixed32_field(6)
    target_game_server_port: int = betterproto.uint32_field(7)


@dataclass(eq=False, repr=False)
class ReportAbuseResponse(betterproto.Message):
    target_steam_id: int = betterproto.fixed64_field(1)
    result: int = betterproto.uint32_field(2)
    error_message: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class NameItemNotification(betterproto.Message):
    player_steamid: int = betterproto.fixed64_field(1)
    item_def_index: int = betterproto.uint32_field(2)
    item_name_custom: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class ClientDisplayNotification(betterproto.Message):
    notification_title_localization_key: str = betterproto.string_field(1)
    notification_body_localization_key: str = betterproto.string_field(2)
    body_substring_keys: list[str] = betterproto.string_field(3)
    body_substring_values: list[str] = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class ShowItemsPickedUp(betterproto.Message):
    player_steamid: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class IncrementKillCountResponse(betterproto.Message):
    killer_account_id: int = betterproto.uint32_field(1)
    num_kills: int = betterproto.uint32_field(2)
    item_def: int = betterproto.uint32_field(3)
    level_type: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class ItemDropRateBonus(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    expiration_date: int = betterproto.fixed32_field(2)
    bonus: float = betterproto.float_field(3)
    bonus_count: int = betterproto.uint32_field(4)
    item_id: int = betterproto.uint64_field(5)
    def_index: int = betterproto.uint32_field(6)


@dataclass(eq=False, repr=False)
class ItemLeagueViewPass(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    league_id: int = betterproto.uint32_field(2)
    admin: int = betterproto.uint32_field(3)
    itemindex: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class ItemEventTicket(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    event_id: int = betterproto.uint32_field(2)
    item_id: int = betterproto.uint64_field(3)


@dataclass(eq=False, repr=False)
class ItemPreviewItemBoughtNotification(betterproto.Message):
    item_def_index: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class StorePurchaseCancel(betterproto.Message):
    txn_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class StorePurchaseCancelResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class StorePurchaseFinalize(betterproto.Message):
    txn_id: int = betterproto.uint64_field(1)


@dataclass(eq=False, repr=False)
class StorePurchaseFinalizeResponse(betterproto.Message):
    result: int = betterproto.uint32_field(1)
    item_ids: list[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class BannedWordListRequest(betterproto.Message):
    ban_list_group_id: int = betterproto.uint32_field(1)
    word_id: int = betterproto.uint32_field(2)


@dataclass(eq=False, repr=False)
class RequestAnnouncements(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class RequestAnnouncementsResponse(betterproto.Message):
    announcement_title: str = betterproto.string_field(1)
    announcement: str = betterproto.string_field(2)
    nextmatch_title: str = betterproto.string_field(3)
    nextmatch: str = betterproto.string_field(4)


@dataclass(eq=False, repr=False)
class BannedWord(betterproto.Message):
    word_id: int = betterproto.uint32_field(1)
    word_type: "GcBannedWordType" = betterproto.enum_field(2)
    word: str = betterproto.string_field(3)


@dataclass(eq=False, repr=False)
class BannedWordListResponse(betterproto.Message):
    ban_list_group_id: int = betterproto.uint32_field(1)
    word_list: "list[BannedWord]" = betterproto.message_field(2)


@dataclass(eq=False, repr=False)
class ToGcBannedWordListBroadcast(betterproto.Message):
    broadcast: "BannedWordListResponse" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class ToGcBannedWordListUpdated(betterproto.Message):
    group_id: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class DefaultEquippedDefinitionInstanceClient(betterproto.Message):
    account_id: int = betterproto.uint32_field(1)
    item_definition: int = betterproto.uint32_field(2)
    class_id: int = betterproto.uint32_field(3)
    slot_id: int = betterproto.uint32_field(4)


@dataclass(eq=False, repr=False)
class ToGcDirtySdoCache(betterproto.Message):
    sdo_type: int = betterproto.uint32_field(1)
    key_uint64: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class ToGcDirtyMultipleSdoCache(betterproto.Message):
    sdo_type: int = betterproto.uint32_field(1)
    key_uint64: list[int] = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class CollectItem(betterproto.Message):
    collection_item_id: int = betterproto.uint64_field(1)
    subject_item_id: int = betterproto.uint64_field(2)


@dataclass(eq=False, repr=False)
class SdoNoMemcached(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ToGcUpdateSqlKeyValue(betterproto.Message):
    key_name: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ToGcIsTrustedServer(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)


@dataclass(eq=False, repr=False)
class ToGcIsTrustedServerResponse(betterproto.Message):
    is_trusted: bool = betterproto.bool_field(1)


@dataclass(eq=False, repr=False)
class ToGcBroadcastConsoleCommand(betterproto.Message):
    con_command: str = betterproto.string_field(1)


@dataclass(eq=False, repr=False)
class ServerVersionUpdated(betterproto.Message):
    server_version: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ClientVersionUpdated(betterproto.Message):
    client_version: int = betterproto.uint32_field(1)


@dataclass(eq=False, repr=False)
class ToGcWebApiAccountChanged(betterproto.Message):
    pass


@dataclass(eq=False, repr=False)
class ToGcRequestPassportItemGrant(betterproto.Message):
    steam_id: int = betterproto.fixed64_field(1)
    league_id: int = betterproto.uint32_field(2)
    reward_flag: int = betterproto.int32_field(3)


@dataclass(eq=False, repr=False)
class GameServerInfo(betterproto.Message):
    server_public_ip_addr: int = betterproto.fixed32_field(1)
    server_private_ip_addr: int = betterproto.fixed32_field(2)
    server_port: int = betterproto.uint32_field(3)
    server_tv_port: int = betterproto.uint32_field(4)
    server_key: str = betterproto.string_field(5)
    server_hibernation: bool = betterproto.bool_field(6)
    server_type: "GameServerInfoServerType" = betterproto.enum_field(7)
    server_region: int = betterproto.uint32_field(8)
    server_loadavg: float = betterproto.float_field(9)
    server_tv_broadcast_time: float = betterproto.float_field(10)
    server_game_time: float = betterproto.float_field(11)
    server_relay_connected_steam_id: int = betterproto.fixed64_field(12)
    relay_slots_max: int = betterproto.uint32_field(13)
    relays_connected: int = betterproto.int32_field(14)
    relay_clients_connected: int = betterproto.int32_field(15)
    relayed_game_server_steam_id: int = betterproto.fixed64_field(16)
    parent_relay_count: int = betterproto.uint32_field(17)
    tv_secret_code: int = betterproto.fixed64_field(18)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# sources: steammessages_gameservers.proto
# plugin: python-betterproto
# Last updated 18/05/2022

from dataclasses import dataclass

import betterproto

from .base import CMsgIpAddress
from .msg import UnifiedMessage


class EQueryType(betterproto.Enum):
    Invalid = 0
    Ping = 1
    Players = 2
    Rules = 3


class GetServerListRequest(UnifiedMessage, um_name="GameServers.GetServerList"):
    filter: str = betterproto.string_field(1)
    limit: int = betterproto.uint32_field(2)


class GetServerListResponse(UnifiedMessage, um_name="GameServers.GetServerList"):
    servers: "list[GetServerListResponseServer]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class GetServerListResponseServer(betterproto.Message):
    addr: str = betterproto.string_field(1)
    gameport: int = betterproto.uint32_field(2)
    specport: int = betterproto.uint32_field(3)
    steamid: int = betterproto.fixed64_field(4)
    name: str = betterproto.string_field(5)
    appid: int = betterproto.uint32_field(6)
    gamedir: str = betterproto.string_field(7)
    version: str = betterproto.string_field(8)
    product: str = betterproto.string_field(9)
    region: int = betterproto.int32_field(10)
    players: int = betterproto.int32_field(11)
    max_players: int = betterproto.int32_field(12)
    bots: int = betterproto.int32_field(13)
    map: str = betterproto.string_field(14)
    secure: bool = betterproto.bool_field(15)
    dedicated: bool = betterproto.bool_field(16)
    os: str = betterproto.string_field(17)
    gametype: str = betterproto.string_field(18)


class GetServerSteamIDsByIpRequest(UnifiedMessage, um_name="GameServers.GetServerSteamIDsByIP"):
    server_ips: list[str] = betterproto.string_field(1)


class IPsWithSteamIDsResponse(UnifiedMessage, um_name="GameServers.GetServerSteamIDsByIP"):
    servers: "list[IPsWithSteamIDsResponseServer]" = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class IPsWithSteamIDsResponseServer(betterproto.Message):
    addr: str = betterproto.string_field(1)
    steamid: int = betterproto.fixed64_field(2)


class GetServerIPsBySteamIdRequest(UnifiedMessage, um_name="GameServers.GetServerIPsBySteamID"):
    server_steamids: list[int] = betterproto.fixed64_field(1)


class QueryRequest(UnifiedMessage, um_name="GameServers.QueryByFakeIP"):
    fake_ip: int = betterproto.uint32_field(1)
    fake_port: int = betterproto.uint32_field(2)
    app_id: int = betterproto.uint32_field(3)
    query_type: EQueryType = betterproto.enum_field(4)


@dataclass(eq=False, repr=False)
class PingQueryData(betterproto.Message):
    server_ip: CMsgIpAddress = betterproto.message_field(1)
    query_port: int = betterproto.uint32_field(2)
    game_port: int = betterproto.uint32_field(3)
    spectator_port: int = betterproto.uint32_field(4)
    spectator_server_name: str = betterproto.string_field(5)
    server_name: str = betterproto.string_field(6)
    steamid: int = betterproto.fixed64_field(7)
    app_id: int = betterproto.uint32_field(8)
    gamedir: str = betterproto.string_field(9)
    map: str = betterproto.string_field(10)
    game_description: str = betterproto.string_field(11)
    gametype: str = betterproto.string_field(12)
    num_players: int = betterproto.uint32_field(13)
    max_players: int = betterproto.uint32_field(14)
    num_bots: int = betterproto.uint32_field(15)
    password: bool = betterproto.bool_field(16)
    secure: bool = betterproto.bool_field(17)
    dedicated: bool = betterproto.bool_field(18)
    version: str = betterproto.string_field(19)
    sdr_popid: int = betterproto.fixed32_field(20)
    sdr_location_string: str = betterproto.string_field(21)


@dataclass(eq=False, repr=False)
class PlayersQueryDataPlayer(betterproto.Message):
    name: str = betterproto.string_field(1)
    score: int = betterproto.uint32_field(2)
    time_played: int = betterproto.uint32_field(3)


@dataclass(eq=False, repr=False)
class PlayersQueryData(betterproto.Message):
    players: list[PlayersQueryDataPlayer] = betterproto.message_field(1)


@dataclass(eq=False, repr=False)
class RulesQueryData(betterproto.Message):
    rules: dict[str, str] = betterproto.map_field(1, betterproto.TYPE_STRING, betterproto.TYPE_STRING)


class QueryResponse(UnifiedMessage, um_name="GameServers.QueryByFakeIP"):
    ping_data: PingQueryData = betterproto.message_field(1)
    players_data: PlayersQueryData = betterproto.message_field(2)
    rules_data: RulesQueryData = betterproto.message_field(3)
